from __future__ import annotations

from datetime import date, datetime, timedelta
import logging
import functools
import itertools
import typing as t
from collections import deque, defaultdict
from functools import reduce, wraps

import sqlglot
from sqlglot import Dialect, exp
from sqlglot.helper import first, merge_ranges, while_changing
from sqlglot.optimizer.annotate_types import TypeAnnotator
from sqlglot.optimizer.scope import find_all_in_scope, walk_in_scope
from sqlglot.schema import ensure_schema


if t.TYPE_CHECKING:
    from dateutil.relativedelta import relativedelta
    from sqlglot.dialects.dialect import DialectType
    from typing_extensions import TypeIs

    DateRange = tuple[date, date]
    DateTruncBinaryTransform = t.Callable[
        [exp.Expr, date, str, Dialect, exp.DataType], t.Optional[exp.Expr]
    ]

logger = logging.getLogger("sqlglot")


# Final means that an expression should not be simplified
FINAL = "final"

SIMPLIFIABLE = (
    exp.Binary,
    exp.Func,
    exp.Lambda,
    exp.Predicate,
    exp.Unary,
)


def simplify(
    expression: exp.Expr,
    constant_propagation: bool = False,
    coalesce_simplification: bool = False,
    dialect: DialectType = None,
):
    """
    Rewrite sqlglot AST to simplify expressions.

    Example:
        >>> import sqlglot
        >>> expression = sqlglot.parse_one("TRUE AND TRUE")
        >>> simplify(expression).sql()
        'TRUE'

    Args:
        expression: expression to simplify
        constant_propagation: whether the constant propagation rule should be used
        coalesce_simplification: whether the simplify coalesce rule should be used.
            This rule tries to remove coalesce functions, which can be useful in certain analyses but
            can leave the query more verbose.
    Returns:
        sqlglot.Expr: simplified expression
    """
    return Simplifier(dialect=dialect).simplify(
        expression,
        constant_propagation=constant_propagation,
        coalesce_simplification=coalesce_simplification,
    )


class UnsupportedUnit(Exception):
    pass


def catch(*exceptions):
    """Decorator that ignores a simplification function if any of `exceptions` are raised"""

    def decorator(func):
        def wrapped(self, expression, *args, **kwargs):
            try:
                return func(self, expression, *args, **kwargs)
            except exceptions:
                return expression

        return wrapped

    return decorator


def annotate_types_on_change(func):
    @wraps(func)
    def _func(self, expression: exp.Expr, *args, **kwargs) -> exp.Expr | None:
        new_expression: exp.Expr | None = func(self, expression, *args, **kwargs)

        if new_expression is None:
            return new_expression

        if self.annotate_new_expressions and expression != new_expression:
            self._annotator.clear()

            # We annotate this to ensure new children nodes are also annotated
            new_expression = self._annotator.annotate(
                expression=new_expression,
                annotate_scope=False,
            )

            # Whatever expression the original expression is transformed into needs to preserve
            # the original type, otherwise the simplification could result in a different schema
            new_expression.type = expression.type

        return new_expression

    return _func


def flatten(expression: exp.Expr) -> exp.Expr:
    """
    A AND (B AND C) -> A AND B AND C
    A OR (B OR C) -> A OR B OR C
    """
    if isinstance(expression, exp.Connector):
        for node in expression.args.values():
            child = node.unnest()
            if isinstance(child, expression.__class__):
                node.replace(child)
    return expression


def simplify_parens(expression: exp.Expr, dialect: DialectType) -> exp.Expr:
    if not isinstance(expression, exp.Paren):
        return expression

    this = expression.this
    parent = expression.parent
    parent_is_predicate = isinstance(parent, exp.Predicate)

    if isinstance(this, exp.Select):
        return expression

    if isinstance(parent, (exp.SubqueryPredicate, exp.Bracket)):
        return expression

    if (
        Dialect.get_or_raise(dialect).REQUIRES_PARENTHESIZED_STRUCT_ACCESS
        and isinstance(parent, exp.Dot)
        and (isinstance(parent.right, (exp.Identifier, exp.Star)))
    ):
        return expression

    if isinstance(this, exp.Predicate) and (
        not (
            parent_is_predicate
            # unary operators that bind tighter than the predicate, unlike NOT
            or isinstance(parent, (exp.Neg, exp.BitwiseNot))
            or (isinstance(parent, exp.Binary) and not isinstance(parent, exp.Connector))
        )
    ):
        return this

    if (
        not isinstance(parent, (exp.Condition, exp.Binary))
        or isinstance(parent, exp.Paren)
        or (
            not isinstance(this, exp.Binary)
            and not (isinstance(this, (exp.Not, exp.Is)) and parent_is_predicate)
        )
        or (isinstance(this, exp.Add) and isinstance(parent, exp.Add))
        or (isinstance(this, exp.Mul) and isinstance(parent, exp.Mul))
        or (isinstance(this, exp.Mul) and isinstance(parent, (exp.Add, exp.Sub)))
    ):
        return this

    return expression


def propagate_constants(expression, root=True):
    """
    Propagate constants for conjunctions in DNF:

    SELECT * FROM t WHERE a = b AND b = 5 becomes
    SELECT * FROM t WHERE a = 5 AND b = 5

    Reference: https://www.sqlite.org/optoverview.html
    """

    if (
        isinstance(expression, exp.And)
        and (root or not expression.same_parent)
        and sqlglot.optimizer.normalize.normalized(expression, dnf=True)
    ):
        constant_mapping = {}
        for expr in walk_in_scope(expression, prune=lambda node: isinstance(node, exp.If)):
            if isinstance(expr, exp.EQ):
                l, r = expr.left, expr.right

                # TODO: create a helper that can be used to detect nested literal expressions such
                # as CAST(123456 AS BIGINT), since we usually want to treat those as literals too
                if isinstance(l, exp.Column) and isinstance(r, exp.Literal):
                    constant_mapping[l] = (id(l), r)

        if constant_mapping:
            for column in find_all_in_scope(expression, exp.Column):
                parent = column.parent
                column_id, constant = constant_mapping.get(column) or (None, None)
                if (
                    column_id is not None
                    and id(column) != column_id
                    and not (isinstance(parent, exp.Is) and isinstance(parent.expression, exp.Null))
                ):
                    column.replace(constant.copy())

    return expression


def _is_number(expression: exp.Expr) -> bool:
    return expression.is_number


def _is_interval(expression: exp.Expr) -> bool:
    return isinstance(expression, exp.Interval) and extract_interval(expression) is not None


def _is_nonnull_constant(expression: exp.Expr) -> bool:
    return isinstance(expression, exp.NONNULL_CONSTANTS) or _is_date_literal(expression)


def _is_constant(expression: exp.Expr) -> bool:
    expr = expression.this if isinstance(expression, exp.Neg) else expression
    return isinstance(expr, exp.CONSTANTS) or _is_date_literal(expr)


def _datetrunc_range(date: date, unit: str, dialect: Dialect) -> DateRange | None:
    """
    Get the date range for a DATE_TRUNC equality comparison:

    Example:
        _datetrunc_range(date(2021-01-01), 'year') == (date(2021-01-01), date(2022-01-01))
    Returns:
        tuple of [min, max) or None if a value can never be equal to `date` for `unit`
    """
    floor = datetime_floor(date, unit, dialect)

    if date != floor:
        # This will always be False, except for NULL values.
        return None

    return floor, floor + interval(unit)


def _datetrunc_eq_expression(
    left: exp.Expr, drange: DateRange, target_type: exp.DataType | None
) -> exp.Expr:
    """Get the logical expression for a date range"""
    return exp.and_(
        left >= date_literal(drange[0], target_type),
        left < date_literal(drange[1], target_type),
        copy=False,
    )


def _datetrunc_eq(
    left: exp.Expr,
    date: date,
    unit: str,
    dialect: Dialect,
    target_type: exp.DataType | None,
) -> exp.Expr | None:
    drange = _datetrunc_range(date, unit, dialect)
    if not drange:
        return None

    return _datetrunc_eq_expression(left, drange, target_type)


def _datetrunc_neq(
    left: exp.Expr,
    date: date,
    unit: str,
    dialect: Dialect,
    target_type: exp.DataType | None,
) -> exp.Expr | None:
    drange = _datetrunc_range(date, unit, dialect)
    if not drange:
        return None

    return exp.or_(
        left < date_literal(drange[0], target_type),
        left >= date_literal(drange[1], target_type),
        copy=False,
    )


def _parenthesize_nested_connector(expression: exp.Expr, parent: exp.Expr | None) -> exp.Expr:
    """
    The generator flattens nested connectors and relies on Paren nodes for grouping.
    Operator precedence varies across dialects, so wrap unless the parent is the same
    connector type, in which case flattening is safe by associativity.
    """
    if isinstance(expression, exp.Connector) and (
        isinstance(parent, exp.Not)
        or (isinstance(parent, exp.Connector) and type(parent) is not type(expression))
    ):
        return exp.paren(expression, copy=False)

    return expression


def always_true(expression: object) -> bool:
    return (isinstance(expression, exp.Boolean) and expression.this) or (
        isinstance(expression, exp.Literal) and expression.is_number and not is_zero(expression)
    )


def always_false(expression: object) -> bool:
    return is_false(expression) or is_null(expression) or is_zero(expression)


def is_zero(expression: object) -> bool:
    return isinstance(expression, exp.Literal) and expression.to_py() == 0


def is_complement(a: object, b: object) -> bool:
    return isinstance(b, exp.Not) and b.this == a


def is_false(a: object) -> bool:
    return type(a) is exp.Boolean and not a.this


def is_null(a: object) -> bool:
    return type(a) is exp.Null


class SupportsComparison(t.Protocol):
    """Protocol for expressions or values that can be compared using <, <=, >, >=."""

    def __lt__(self, other: t.Any) -> bool: ...
    def __le__(self, other: t.Any) -> bool: ...
    def __gt__(self, other: t.Any) -> bool: ...
    def __ge__(self, other: t.Any) -> bool: ...


def eval_boolean(
    expression: object, a: SupportsComparison, b: SupportsComparison
) -> exp.Boolean | None:
    if isinstance(expression, (exp.EQ, exp.Is)):
        if isinstance(expression, exp.Is) and expression.args.get("negate"):
            return boolean_literal(a != b)
        return boolean_literal(a == b)
    if isinstance(expression, exp.NEQ):
        return boolean_literal(a != b)
    if isinstance(expression, exp.GT):
        return boolean_literal(a > b)
    if isinstance(expression, exp.GTE):
        return boolean_literal(a >= b)
    if isinstance(expression, exp.LT):
        return boolean_literal(a < b)
    if isinstance(expression, exp.LTE):
        return boolean_literal(a <= b)
    return None


def cast_as_date(value: datetime | date | str) -> date | None:
    if isinstance(value, datetime):
        return value.date()
    if isinstance(value, date):
        return value
    try:
        return datetime.fromisoformat(value).date()
    except ValueError:
        return None


def cast_as_datetime(
    value: datetime | date | str,
) -> datetime | None:
    if isinstance(value, datetime):
        return value
    if isinstance(value, date):
        return datetime(year=value.year, month=value.month, day=value.day)
    try:
        return datetime.fromisoformat(value)
    except ValueError:
        return None


def cast_value(value: datetime | date | str, to: exp.DataType) -> date | date | None:
    if not value:
        return None
    if to.is_type(exp.DType.DATE):
        return cast_as_date(value)
    if to.is_type(*exp.DataType.TEMPORAL_TYPES):
        return cast_as_datetime(value)
    return None


def extract_date(cast: exp.Expr) -> date | date | None:
    if isinstance(cast, exp.Cast):
        to = cast.to
    elif isinstance(cast, exp.TsOrDsToDate) and not cast.args.get("format"):
        to = exp.DType.DATE.into_expr()
    else:
        return None

    if isinstance(cast.this, exp.Literal):
        value: t.Any = cast.this.name
    elif isinstance(cast.this, (exp.Cast, exp.TsOrDsToDate)):
        value = extract_date(cast.this)
    else:
        return None
    return cast_value(value, to)


def _is_date_literal(expression: exp.Expr) -> bool:
    return extract_date(expression) is not None


def extract_interval(expression: exp.Expr) -> relativedelta | None:
    try:
        n = int(expression.this.to_py())
        unit = expression.text("unit").lower()
        return interval(unit, n)
    except (UnsupportedUnit, ModuleNotFoundError, ValueError):
        return None


def extract_type(*expressions: exp.Expr):
    target_type = None
    for expression in expressions:
        target_type = expression.to if isinstance(expression, exp.Cast) else expression.type
        if target_type:
            break

    return target_type


def date_literal(date: object, target_type=None) -> exp.Expr:
    if not target_type or not target_type.is_type(*exp.DataType.TEMPORAL_TYPES):
        target_type = exp.DType.DATETIME if isinstance(date, datetime) else exp.DType.DATE

    return exp.cast(exp.Literal.string(date), target_type)


def interval(unit: str, n: int = 1) -> relativedelta:
    from dateutil.relativedelta import relativedelta

    if unit == "year":
        return relativedelta(years=1 * n)
    if unit == "quarter":
        return relativedelta(months=3 * n)
    if unit == "month":
        return relativedelta(months=1 * n)
    if unit == "week":
        return relativedelta(weeks=1 * n)
    if unit == "day":
        return relativedelta(days=1 * n)
    if unit == "hour":
        return relativedelta(hours=1 * n)
    if unit == "minute":
        return relativedelta(minutes=1 * n)
    if unit == "second":
        return relativedelta(seconds=1 * n)
    if unit == "millisecond":
        return relativedelta(microseconds=1000 * n)
    if unit == "microsecond":
        return relativedelta(microseconds=1 * n)

    raise UnsupportedUnit(f"Unsupported unit: {unit}")


def datetime_floor(d: date, unit: str, dialect: Dialect) -> date:
    # Truncate sub-day units — only valid for datetime inputs
    if isinstance(d, datetime):
        if unit == "hour":
            return d.replace(minute=0, second=0, microsecond=0)
        elif unit == "minute":
            return d.replace(second=0, microsecond=0)
        elif unit == "second":
            return d.replace(microsecond=0)
        elif unit == "millisecond":
            return d.replace(microsecond=(d.microsecond // 1000) * 1000)
        elif unit == "microsecond":
            return d

    # Truncate date-level units, shared for both date and datetime
    if unit == "year":
        result: date = d.replace(month=1, day=1)
    elif unit == "quarter":
        if d.month <= 3:
            result = d.replace(month=1, day=1)
        elif d.month <= 6:
            result = d.replace(month=4, day=1)
        elif d.month <= 9:
            result = d.replace(month=7, day=1)
        else:
            result = d.replace(month=10, day=1)
    elif unit == "month":
        result = d.replace(month=d.month, day=1)
    elif unit == "week":
        # Week truncation respects dialect.WEEK_OFFSET (0=Monday, -1=Sunday)
        result = d - timedelta(days=(d.weekday() - dialect.WEEK_OFFSET) % 7)
    elif unit == "day":
        result = d
    else:
        raise UnsupportedUnit(f"Unsupported unit: {unit}")

    # For datetime inputs, zero out the time component after date-level truncation
    if isinstance(result, datetime):
        return result.replace(hour=0, minute=0, second=0, microsecond=0)
    return result


def _trunc_unit(unit: exp.Expr, dialect: Dialect) -> str:
    if isinstance(unit, exp.WeekStart):
        from sqlglot.dialects.dialect import WEEK_START_DAY_TO_DOW, week_offset_to_dow

        if WEEK_START_DAY_TO_DOW.get(unit.name.upper()) != week_offset_to_dow(dialect.WEEK_OFFSET):
            raise UnsupportedUnit(f"Unsupported unit: {unit}")
        return "week"

    return unit.name.lower()


def date_ceil(d: date, unit: str, dialect: Dialect) -> date:
    floor = datetime_floor(d, unit, dialect)

    if floor == d:
        return d

    return floor + interval(unit)


def boolean_literal(condition: bool | exp.Predicate) -> exp.Boolean:
    return exp.true() if condition else exp.false()


class Simplifier:
    def __init__(self, dialect: DialectType = None, annotate_new_expressions: bool = True):
        self.dialect = Dialect.get_or_raise(dialect)
        self.annotate_new_expressions = annotate_new_expressions

        self._annotator: TypeAnnotator = TypeAnnotator(
            schema=ensure_schema(None, dialect=self.dialect), overwrite_types=False
        )

    # Value ranges for byte-sized signed/unsigned integers
    TINYINT_MIN: t.ClassVar = -128
    TINYINT_MAX: t.ClassVar = 127
    UTINYINT_MIN: t.ClassVar = 0
    UTINYINT_MAX: t.ClassVar = 255

    COMPLEMENT_COMPARISONS: t.ClassVar = {
        exp.LT: exp.GTE,
        exp.GT: exp.LTE,
        exp.LTE: exp.GT,
        exp.GTE: exp.LT,
        exp.EQ: exp.NEQ,
        exp.NEQ: exp.EQ,
    }

    COMPLEMENT_SUBQUERY_PREDICATES: t.ClassVar = {
        exp.All: exp.Any,
        exp.Any: exp.All,
    }

    LT_LTE: t.ClassVar = (exp.LT, exp.LTE)
    GT_GTE: t.ClassVar = (exp.GT, exp.GTE)

    COMPARISONS: t.ClassVar = (
        *LT_LTE,
        *GT_GTE,
        exp.EQ,
        exp.NEQ,
        exp.Is,
    )

    # Operand types that allow a connector pair to combine in _flat_simplify: constants
    # (matched by the is_false/is_null/is_zero/always_true checks) and comparisons.
    CONNECTOR_COMBINABLE: t.ClassVar = (exp.Boolean, exp.Literal, exp.Null, *COMPARISONS)

    INVERSE_COMPARISONS: t.ClassVar[dict[type[exp.Expr], type[exp.Expr]]] = {
        exp.LT: exp.GT,
        exp.GT: exp.LT,
        exp.LTE: exp.GTE,
        exp.GTE: exp.LTE,
    }

    NONDETERMINISTIC: t.ClassVar = (exp.Rand, exp.Randn)
    AND_OR: t.ClassVar = (exp.And, exp.Or)

    INVERSE_DATE_OPS: t.ClassVar[dict[type[exp.Expr], type[exp.Expr]]] = {
        exp.DateAdd: exp.Sub,
        exp.DateSub: exp.Add,
        exp.DatetimeAdd: exp.Sub,
        exp.DatetimeSub: exp.Add,
    }

    INVERSE_OPS: t.ClassVar[dict[type[exp.Expr], type[exp.Expr]]] = {
        **INVERSE_DATE_OPS,
        exp.Add: exp.Sub,
        exp.Sub: exp.Add,
    }

    NULL_OK: t.ClassVar = (exp.NullSafeEQ, exp.NullSafeNEQ, exp.PropertyEQ)

    CONCATS: t.ClassVar = (exp.Concat, exp.DPipe)

    DATETRUNC_BINARY_COMPARISONS: t.ClassVar[dict[type[exp.Expr], DateTruncBinaryTransform]] = {
        exp.LT: lambda l, dt, u, d, t: (
            l
            < date_literal(
                dt if dt == datetime_floor(dt, u, d) else datetime_floor(dt, u, d) + interval(u),
                t,
            )
        ),
        exp.GT: lambda l, dt, u, d, t: l >= date_literal(datetime_floor(dt, u, d) + interval(u), t),
        exp.LTE: lambda l, dt, u, d, t: l < date_literal(datetime_floor(dt, u, d) + interval(u), t),
        exp.GTE: lambda l, dt, u, d, t: l >= date_literal(date_ceil(dt, u, d), t),
        exp.EQ: _datetrunc_eq,
        exp.NEQ: _datetrunc_neq,
    }

    DATETRUNC_COMPARISONS: t.ClassVar = {exp.In, *DATETRUNC_BINARY_COMPARISONS}
    DATETRUNCS: t.ClassVar = (exp.DateTrunc, exp.TimestampTrunc)

    SAFE_CONNECTOR_ELIMINATION_RESULT: t.ClassVar = (exp.Connector, exp.Boolean)

    # CROSS joins result in an empty table if the right table is empty.
    # So we can only simplify certain types of joins to CROSS.
    # Or in other words, LEFT JOIN x ON TRUE != CROSS JOIN x
    JOINS: t.ClassVar = {
        ("", ""),
        ("", "INNER"),
        ("RIGHT", ""),
        ("RIGHT", "OUTER"),
    }

    def simplify(
        self,
        expression: exp.Expr,
        constant_propagation: bool = False,
        coalesce_simplification: bool = False,
    ) -> exp.Expr:
        wheres: list[exp.Where] = []
        joins: list[exp.Join] = []

        for node in expression.walk(
            prune=lambda n: bool(isinstance(n, exp.Condition) or n.meta_get(FINAL))
        ):
            if node.meta_get(FINAL):
                continue

            # group by expressions cannot be simplified, for example
            # select x + 1 + 1 FROM y GROUP BY x + 1 + 1
            # the projection must exactly match the group by key
            group = node.args.get("group")

            if group and hasattr(node, "selects"):
                groups = set(group.expressions)
                group.meta[FINAL] = True

                for s in node.selects:
                    for n in s.walk():
                        if n in groups:
                            s.meta[FINAL] = True
                            break

                having = node.args.get("having")

                if having:
                    for n in having.walk():
                        if n in groups:
                            having.meta[FINAL] = True
                            break

            if isinstance(node, exp.Condition):
                simplified = while_changing(
                    node, lambda e: self._simplify(e, constant_propagation, coalesce_simplification)
                )

                if node is expression:
                    expression = simplified
            elif isinstance(node, exp.Where):
                wheres.append(node)
            elif isinstance(node, exp.Join):
                # snowflake match_conditions have very strict ordering rules
                if match := node.args.get("match_condition"):
                    match.meta[FINAL] = True

                joins.append(node)

        for where in wheres:
            if always_true(where.this):
                where.pop()
        for join in joins:
            if (
                always_true(join.args.get("on"))
                and not join.args.get("using")
                and not join.args.get("method")
                and (join.side, join.kind) in self.JOINS
            ):
                join.args["on"].pop()
                join.set("side", None)
                join.set("kind", "CROSS")

        return expression

    def _simplify(
        self, expression: exp.Expr, constant_propagation: bool, coalesce_simplification: bool
    ):
        pre_transformation_stack = [expression]
        post_transformation_stack = []

        while pre_transformation_stack:
            original = pre_transformation_stack.pop()
            node = original

            if not isinstance(node, SIMPLIFIABLE):
                if isinstance(node, exp.Query):
                    self.simplify(node, constant_propagation, coalesce_simplification)
                continue

            parent = node.parent
            root = node is expression

            node = self.rewrite_between(node)
            node = self.uniq_sort(node, root)
            node = self.absorb_and_eliminate(node, root)
            node = self.simplify_concat(node)
            node = self.simplify_conditionals(node)

            if constant_propagation:
                node = propagate_constants(node, root)

            if node is not original:
                original.replace(node)

            pre_transformation_stack.extend(
                n for n in node.iter_expressions(reverse=True) if not n.meta_get(FINAL)
            )
            post_transformation_stack.append((node, parent))

        while post_transformation_stack:
            original, parent = post_transformation_stack.pop()
            root = original is expression

            # Resets parent, arg_key, index pointers– this is needed because some of the
            # previous transformations mutate the AST, leading to an inconsistent state.
            # We only fix pointers instead of calling `set` because the values are unchanged:
            # actual mutations go through `set`/`replace`, which already invalidate cached
            # hashes, so clearing them again here would force `while_changing` to rehash
            # entire subtrees after every (mostly no-op) pass.
            for k, v in tuple(original.args.items()):
                if v is None:
                    original.args.pop(k)
                else:
                    original._set_parent(k, v)

            # Post-order transformations
            node = self.simplify_not(original)
            node = flatten(node)
            node = self.simplify_connectors(node, root)
            node = self.remove_complements(node, root)

            if coalesce_simplification:
                node = self.simplify_coalesce(node)
            node.parent = parent

            node = self.simplify_literals(node, root)
            node = self.simplify_equality(node)
            node = simplify_parens(node, dialect=self.dialect)
            node = self.simplify_datetrunc(node)
            node = self.sort_comparison(node)
            node = self.simplify_startswith(node)

            if node is not original:
                original.replace(node)

        return node

    @annotate_types_on_change
    def rewrite_between(self, expression: exp.Expr) -> exp.Expr:
        """Rewrite x between y and z to x >= y AND x <= z.

        This is done because comparison simplification is only done on lt/lte/gt/gte.
        """
        if isinstance(expression, exp.Between):
            negate = isinstance(expression.parent, exp.Not)

            expression = exp.and_(
                exp.GTE(this=expression.this.copy(), expression=expression.args["low"]),
                exp.LTE(this=expression.this.copy(), expression=expression.args["high"]),
                copy=False,
            )

            if negate:
                expression = exp.paren(expression, copy=False)

        return expression

    @annotate_types_on_change
    def simplify_not(self, expression: exp.Expr) -> exp.Expr:
        """
        Demorgan's Law
        NOT (x OR y) -> NOT x AND NOT y
        NOT (x AND y) -> NOT x OR NOT y
        """
        if isinstance(expression, exp.Not):
            this = expression.this
            if is_null(this):
                return _parenthesize_nested_connector(
                    exp.and_(exp.null(), exp.true(), copy=False), expression.parent
                )
            if this.__class__ in self.COMPLEMENT_COMPARISONS:
                right = this.expression
                complement_subquery_predicate = self.COMPLEMENT_SUBQUERY_PREDICATES.get(
                    right.__class__
                )
                if complement_subquery_predicate:
                    right = complement_subquery_predicate(this=right.this)

                return self.COMPLEMENT_COMPARISONS[this.__class__](this=this.this, expression=right)
            if isinstance(this, exp.Paren):
                condition = this.unnest()
                if isinstance(condition, exp.And):
                    return exp.paren(
                        exp.or_(
                            exp.not_(condition.left, copy=False),
                            exp.not_(condition.right, copy=False),
                            copy=False,
                        ),
                        copy=False,
                    )
                if isinstance(condition, exp.Or):
                    return exp.paren(
                        exp.and_(
                            exp.not_(condition.left, copy=False),
                            exp.not_(condition.right, copy=False),
                            copy=False,
                        ),
                        copy=False,
                    )
                if is_null(condition):
                    return _parenthesize_nested_connector(
                        exp.and_(exp.null(), exp.true(), copy=False), expression.parent
                    )
            if always_true(this):
                return exp.false()
            if is_false(this):
                return exp.true()
            if isinstance(this, exp.Not) and self.dialect.SAFE_TO_ELIMINATE_DOUBLE_NEGATION:
                inner = this.this
                if inner.is_type(exp.DType.BOOLEAN):
                    # double negation
                    # NOT NOT x -> x, if x is BOOLEAN type
                    return inner
        return expression

    @annotate_types_on_change
    def simplify_connectors(self, expression: exp.Expr, root: bool = True) -> exp.Expr:
        def _simplify_connectors(expression: exp.Expr, left: exp.Expr, right: exp.Expr):
            if isinstance(expression, exp.And):
                if is_false(left) or is_false(right):
                    return exp.false()
                if is_zero(left) or is_zero(right):
                    return exp.false()
                if (
                    (is_null(left) and is_null(right))
                    or (is_null(left) and always_true(right))
                    or (always_true(left) and is_null(right))
                ):
                    return exp.null()
                if always_true(left) and always_true(right):
                    return exp.true()
                if always_true(left):
                    return right
                if always_true(right):
                    return left
                return self._simplify_comparison(expression, left, right)
            elif isinstance(expression, exp.Or):
                if always_true(left) or always_true(right):
                    return exp.true()
                if (
                    (is_null(left) and is_null(right))
                    or (is_null(left) and always_false(right))
                    or (always_false(left) and is_null(right))
                ):
                    return exp.null()
                if is_false(left):
                    return right
                if is_false(right):
                    return left
                return self._simplify_comparison(expression, left, right, or_=True)

        if isinstance(expression, exp.Connector):
            original_parent = expression.parent
            expression = self._flat_simplify(expression, _simplify_connectors, root)

            # If we reduced a connector to, e.g., a column (t1 AND ... AND tn -> Tk), then we need
            # to ensure that the resulting type is boolean. We know this is true only for connectors,
            # boolean values and columns that are essentially operands to a connector:
            #
            # A AND (((B)))
            #          ~ this is safe to keep because it will eventually be part of another connector
            if not isinstance(
                expression, self.SAFE_CONNECTOR_ELIMINATION_RESULT
            ) and not expression.is_type(exp.DType.BOOLEAN):
                while True:
                    if isinstance(original_parent, exp.Connector):
                        break
                    if not isinstance(original_parent, exp.Paren):
                        expression = expression.and_(exp.true(), copy=False)
                        break

                    original_parent = original_parent.parent

        return expression

    @annotate_types_on_change
    def _simplify_comparison(
        self, expression: exp.Expr, left: exp.Expr, right: exp.Expr, or_: bool = False
    ) -> exp.Expr | None:
        if isinstance(left, self.COMPARISONS) and isinstance(right, self.COMPARISONS):
            if any(isinstance(e, exp.Is) and e.args.get("negate") for e in (left, right)):
                return None

            ll, lr = left.this, left.expression
            rl, rr = right.this, right.expression

            largs = {ll, lr}
            rargs = {rl, rr}

            matching = largs & rargs
            columns = {
                m for m in matching if not _is_constant(m) and not m.find(*self.NONDETERMINISTIC)
            }

            if matching and columns:
                try:
                    l = first(largs - columns)
                    r = first(rargs - columns)
                except StopIteration:
                    return expression

                if l.is_number and r.is_number:
                    l = l.to_py()
                    r = r.to_py()
                elif l.is_string and r.is_string:
                    l = l.name
                    r = r.name
                else:
                    l = extract_date(l)
                    if not l:
                        return None
                    r = extract_date(r)
                    if not r:
                        return None
                    # python won't compare date and datetime, but many engines will upcast
                    l, r = cast_as_datetime(l), cast_as_datetime(r)

                for (a, av), (b, bv) in itertools.permutations(((left, l), (right, r))):
                    if isinstance(a, self.LT_LTE) and isinstance(b, self.LT_LTE):
                        return left if (av > bv if or_ else av <= bv) else right
                    if isinstance(a, self.GT_GTE) and isinstance(b, self.GT_GTE):
                        return left if (av < bv if or_ else av >= bv) else right

                    # we can't ever shortcut to true because the column could be null
                    if not or_:
                        if isinstance(a, exp.LT) and isinstance(b, self.GT_GTE):
                            if av <= bv:
                                return exp.false()
                        elif isinstance(a, exp.GT) and isinstance(b, self.LT_LTE):
                            if av >= bv:
                                return exp.false()
                        elif isinstance(a, exp.EQ):
                            if isinstance(b, exp.LT):
                                return exp.false() if av >= bv else a
                            if isinstance(b, exp.LTE):
                                return exp.false() if av > bv else a
                            if isinstance(b, exp.GT):
                                return exp.false() if av <= bv else a
                            if isinstance(b, exp.GTE):
                                return exp.false() if av < bv else a
                            if isinstance(b, exp.NEQ):
                                return exp.false() if av == bv else a
        return None

    @annotate_types_on_change
    def remove_complements(self, expression: object, root: bool = True) -> object:
        """
        Removing complements.

        A AND NOT A -> FALSE (only for non-NULL A)
        A OR NOT A -> TRUE (only for non-NULL A)
        """
        if isinstance(expression, self.AND_OR) and (root or not expression.same_parent):
            ops = set(expression.flatten())
            for op in ops:
                if isinstance(op, exp.Not) and op.this in ops:
                    if expression.meta_get("nonnull") is True:
                        return exp.false() if isinstance(expression, exp.And) else exp.true()

        return expression

    @annotate_types_on_change
    def uniq_sort(self, expression: object, root: bool = True) -> object:
        """
        Uniq and sort a connector.

        C AND A AND B AND B -> A AND B AND C
        """
        if isinstance(expression, exp.Connector) and (root or not expression.same_parent):
            flattened = tuple(expression.flatten())

            if isinstance(expression, exp.Xor):
                result_func = exp.xor
                # Do not deduplicate XOR as A XOR A != A if A == True
                deduped = None
                arr = tuple((gen(e), e) for e in flattened)
            else:
                result_func = exp.and_ if isinstance(expression, exp.And) else exp.or_
                deduped = {gen(e): e for e in flattened}
                arr = tuple(deduped.items())

            # check if the operands are already sorted, if not sort them
            # A AND C AND B -> A AND B AND C
            for i, (sql, e) in enumerate(arr[1:]):
                if sql < arr[i][0]:
                    expression = result_func(*(e for _, e in sorted(arr)), copy=False)
                    break
            else:
                # we didn't have to sort but maybe we need to dedup
                if deduped and len(deduped) < len(flattened):
                    unique_operand = flattened[0]
                    if len(deduped) == 1:
                        expression = unique_operand.and_(exp.true(), copy=False)
                    else:
                        expression = result_func(*deduped.values(), copy=False)

        return expression

    @annotate_types_on_change
    def absorb_and_eliminate(self, expression: object, root: bool = True) -> object:
        """
        absorption:
            A AND (A OR B) -> A
            A OR (A AND B) -> A
            A AND (NOT A OR B) -> A AND B (only for non-NULL A)
            A OR (NOT A AND B) -> A OR B (only for non-NULL A)
        elimination:
            (A AND B) OR (A AND NOT B) -> A (only for non-NULL B)
            (A OR B) AND (A OR NOT B) -> A (only for non-NULL B)
        """
        if isinstance(expression, self.AND_OR) and (root or not expression.same_parent):
            kind = exp.Or if isinstance(expression, exp.And) else exp.And

            ops = tuple(expression.flatten())

            # Initialize lookup tables:
            # Set of all operands, used to find complements for absorption.
            op_set = set()
            # Sub-operands, used to find subsets for absorption.
            subops = defaultdict(list)
            # Pairs of complements, used for elimination.
            pairs = defaultdict(list)

            # Populate the lookup tables
            for op in ops:
                op_set.add(op)

                if not isinstance(op, kind):
                    # In cases like: A OR (A AND B)
                    # Subop will be: ^
                    subops[op].append({op})
                    continue

                # In cases like: (A AND B) OR (A AND B AND C)
                # Subops will be: ^     ^
                subset = set(op.flatten())
                for i in subset:
                    subops[i].append(subset)

                a, b = op.unnest_operands()

                if isinstance(a, exp.Not) and a.this.meta_get("nonnull") is True:
                    pairs[frozenset((a.this, b))].append((op, b))
                if isinstance(b, exp.Not) and b.this.meta_get("nonnull") is True:
                    pairs[frozenset((a, b.this))].append((op, a))

            for op in ops:
                if not isinstance(op, kind):
                    continue

                a, b = op.unnest_operands()

                # Absorb
                if (
                    isinstance(a, exp.Not)
                    and a.this in op_set
                    and a.this.meta_get("nonnull") is True
                ):
                    a.replace(exp.true() if kind == exp.And else exp.false())
                    continue
                if (
                    isinstance(b, exp.Not)
                    and b.this in op_set
                    and b.this.meta_get("nonnull") is True
                ):
                    b.replace(exp.true() if kind == exp.And else exp.false())
                    continue
                superset = set(op.flatten())
                if any(any(subset < superset for subset in subops[i]) for i in superset):
                    op.replace(exp.false() if kind == exp.And else exp.true())
                    continue

                # Eliminate
                for other, complement in pairs[frozenset((a, b))]:
                    op.replace(complement)
                    other.replace(complement)

        return expression

    @annotate_types_on_change
    @catch(ModuleNotFoundError, UnsupportedUnit)
    def simplify_equality(self, expression: exp.Expr) -> exp.Expr:
        """
        Use the subtraction and addition properties of equality to simplify expressions:

            x + 1 = 3 becomes x = 2

        There are two binary operations in the above expression: + and =
        Here's how we reference all the operands in the code below:

            l     r
            x + 1 = 3
            a   b

        Subtraction is not commutative, so when the variable is the subtrahend the operands
        can't simply be swapped; the comparison is inverted instead:

            5 - x = 2 becomes x = 3
            5 - x < 2 becomes x > 3
        """
        if isinstance(expression, self.COMPARISONS):
            l, r = expression.left, expression.right

            if l.__class__ not in self.INVERSE_OPS:
                return expression

            if r.is_number:
                a_predicate = _is_number
                b_predicate = _is_number
            elif _is_date_literal(r):
                a_predicate = _is_date_literal
                b_predicate = _is_interval
            else:
                return expression

            if l.__class__ in self.INVERSE_DATE_OPS:
                l = t.cast(exp.IntervalOp, l)
                a: exp.Expr = l.this
                b: exp.Expr = l.interval()
            else:
                l = t.cast(exp.Binary, l)
                a, b = l.left, l.right

            if not a_predicate(a) and b_predicate(b):
                pass
            elif not a_predicate(b) and b_predicate(a):
                if isinstance(l, exp.Sub):
                    return self.INVERSE_COMPARISONS.get(expression.__class__, expression.__class__)(
                        this=b, expression=exp.Sub(this=a, expression=r)
                    )
                a, b = b, a
            else:
                return expression

            return expression.__class__(
                this=a, expression=self.INVERSE_OPS[l.__class__](this=r, expression=b)
            )
        return expression

    def _is_inverse_date_op(self, expression: exp.Expr) -> TypeIs[exp.IntervalOp]:
        return type(expression) in self.INVERSE_DATE_OPS

    @annotate_types_on_change
    def simplify_literals(self, expression: exp.Expr, root: bool = True) -> exp.Expr:
        if isinstance(expression, exp.Binary) and not isinstance(expression, exp.Connector):
            return self._flat_simplify(expression, self._simplify_binary, root)

        if isinstance(expression, exp.Neg) and isinstance(expression.this, exp.Neg):
            return expression.this.this

        if self._is_inverse_date_op(expression):
            return (
                self._simplify_binary(expression, expression.this, expression.interval())
                or expression
            )

        return expression

    def _simplify_integer_cast(self, expr: t.Any) -> t.Any:
        if isinstance(expr, exp.Cast) and isinstance(expr.this, exp.Cast):
            this = self._simplify_integer_cast(expr.this)
        else:
            this = expr.this

        if isinstance(expr, exp.Cast) and this.is_int:
            num = this.to_py()

            # Remove the (up)cast from small (byte-sized) integers in predicates which is side-effect free. Downcasts on any
            # integer type might cause overflow, thus the cast cannot be eliminated and the behavior is
            # engine-dependent
            if (
                self.TINYINT_MIN <= num <= self.TINYINT_MAX
                and expr.to.this in exp.DataType.SIGNED_INTEGER_TYPES
            ) or (
                self.UTINYINT_MIN <= num <= self.UTINYINT_MAX
                and expr.to.this in exp.DataType.UNSIGNED_INTEGER_TYPES
            ):
                return this

        return expr

    def _simplify_binary(self, expression, a, b):
        if isinstance(expression, self.COMPARISONS):
            a = self._simplify_integer_cast(a)
            b = self._simplify_integer_cast(b)

        if isinstance(expression, exp.Is):
            if isinstance(b, exp.Not):
                c = b.this
                not_ = True
            else:
                c = b
                not_ = False

            if expression.args.get("negate"):
                not_ = not not_

            if is_null(c):
                if isinstance(a, exp.Literal):
                    return exp.true() if not_ else exp.false()
                if is_null(a):
                    return exp.false() if not_ else exp.true()
        elif isinstance(expression, self.NULL_OK):
            return None
        elif (is_null(a) or is_null(b)) and isinstance(expression.parent, exp.If):
            return exp.null()

        if a.is_number and b.is_number:
            num_a = a.to_py()
            num_b = b.to_py()

            if isinstance(expression, exp.Add):
                return exp.Literal.number(num_a + num_b)
            if isinstance(expression, exp.Mul):
                return exp.Literal.number(num_a * num_b)

            # We only simplify Sub, Div if a and b have the same parent because they're not associative
            if isinstance(expression, exp.Sub):
                return exp.Literal.number(num_a - num_b) if a.parent is b.parent else None
            if isinstance(expression, exp.Div):
                # engines have differing int div behavior so intdiv is not safe
                if (isinstance(num_a, int) and isinstance(num_b, int)) or a.parent is not b.parent:
                    return None
                return exp.Literal.number(num_a / num_b)

            boolean = eval_boolean(expression, num_a, num_b)

            if boolean:
                return boolean
        elif a.is_string and b.is_string:
            boolean = eval_boolean(expression, a.this, b.this)

            if boolean:
                return boolean
        elif _is_date_literal(a) and isinstance(b, exp.Interval):
            date, b = extract_date(a), extract_interval(b)
            if date and b:
                if isinstance(expression, (exp.Add, exp.DateAdd, exp.DatetimeAdd)):
                    return date_literal(date + b, extract_type(a))
                if isinstance(expression, (exp.Sub, exp.DateSub, exp.DatetimeSub)):
                    return date_literal(date - b, extract_type(a))
        elif isinstance(a, exp.Interval) and _is_date_literal(b):
            a, date = extract_interval(a), extract_date(b)
            # you cannot subtract a date from an interval
            if a and b and isinstance(expression, exp.Add):
                return date_literal(a + date, extract_type(b))
        elif _is_date_literal(a) and _is_date_literal(b):
            if isinstance(expression, exp.Predicate):
                a, b = extract_date(a), extract_date(b)
                boolean = eval_boolean(expression, a, b)
                if boolean:
                    return boolean

        return None

    @annotate_types_on_change
    def simplify_coalesce(self, expression: exp.Expr) -> exp.Expr:
        # COALESCE(x) -> x
        if (
            isinstance(expression, exp.Coalesce)
            and (not expression.expressions or _is_nonnull_constant(expression.this))
            # COALESCE is also used as a Spark partitioning hint
            and not isinstance(expression.parent, exp.Hint)
        ):
            return expression.this

        if self.dialect.COALESCE_COMPARISON_NON_STANDARD:
            return expression

        if not isinstance(expression, self.COMPARISONS):
            return expression

        if isinstance(expression.left, exp.Coalesce):
            coalesce = expression.left
            other = expression.right
        elif isinstance(expression.right, exp.Coalesce):
            coalesce = expression.right
            other = expression.left
        else:
            return expression

        # This transformation is valid for non-constants,
        # but it really only does anything if they are both constants.
        if not _is_constant(other):
            return expression

        # Find the first constant arg
        for arg_index, arg in enumerate(coalesce.expressions):
            if _is_constant(arg):
                break
        else:
            return expression

        coalesce.set("expressions", coalesce.expressions[:arg_index])

        # Remove the COALESCE function. This is an optimization, skipping a simplify iteration,
        # since we already remove COALESCE at the top of this function.
        this: exp.Expr = coalesce if coalesce.expressions else coalesce.this

        # This expression is more complex than when we started, but it will get simplified further
        return exp.paren(
            exp.or_(
                exp.and_(
                    this.is_(exp.null()).not_(copy=False),
                    expression.copy(),
                    copy=False,
                ),
                exp.and_(
                    this.is_(exp.null()),
                    type(expression)(this=arg.copy(), expression=other.copy()),
                    copy=False,
                ),
                copy=False,
            ),
            copy=False,
        )

    @annotate_types_on_change
    def simplify_concat(self, expression):
        """Reduces all groups that contain string literals by concatenating them."""
        if not isinstance(expression, self.CONCATS) or (
            # We can't reduce a CONCAT_WS call if we don't statically know the separator
            isinstance(expression, exp.ConcatWs) and not expression.expressions[0].is_string
        ):
            return expression

        if isinstance(expression, exp.ConcatWs):
            sep_expr, *expressions = expression.expressions
            sep = sep_expr.name
            concat_type = exp.ConcatWs
        else:
            expressions = expression.expressions
            sep = ""
            concat_type = exp.Concat

        args = {
            "safe": expression.args.get("safe"),
            "coalesce": expression.args.get("coalesce"),
        }

        new_args = []
        for is_string_group, group in itertools.groupby(
            expressions or expression.flatten(), lambda e: e.is_string
        ):
            if is_string_group:
                new_args.append(exp.Literal.string(sep.join(string.name for string in group)))
            else:
                new_args.extend(group)

        if len(new_args) == 1 and new_args[0].is_string:
            return new_args[0]

        if concat_type is exp.ConcatWs:
            new_args = [sep_expr] + new_args
        elif isinstance(expression, exp.DPipe):
            return reduce(lambda x, y: exp.DPipe(this=x, expression=y, safe=args["safe"]), new_args)

        return concat_type(expressions=new_args, **args)

    @annotate_types_on_change
    def simplify_conditionals(self, expression):
        """Simplifies expressions like IF, CASE if their condition is statically known."""
        if isinstance(expression, exp.Case):
            this = expression.this
            for case in expression.args["ifs"]:
                cond = case.this
                if this:
                    # Convert CASE x WHEN matching_value ... to CASE WHEN x = matching_value ...
                    cond = cond.replace(this.pop().eq(cond))

                if always_true(cond):
                    return case.args["true"]

                if always_false(cond):
                    case.pop()
                    if not expression.args["ifs"]:
                        return expression.args.get("default") or exp.null()
        elif isinstance(expression, exp.If) and not isinstance(expression.parent, exp.Case):
            if always_true(expression.this):
                return expression.args["true"]
            if always_false(expression.this):
                return expression.args.get("false") or exp.null()

        return expression

    @annotate_types_on_change
    def simplify_startswith(self, expression: exp.Expr) -> exp.Expr:
        """
        Reduces a prefix check to either TRUE or FALSE if both the string and the
        prefix are statically known.

        Example:
            >>> from sqlglot import parse_one
            >>> Simplifier().simplify_startswith(parse_one("STARTSWITH('foo', 'f')")).sql()
            'TRUE'
        """
        if (
            isinstance(expression, exp.StartsWith)
            and expression.this.is_string
            and expression.expression.is_string
        ):
            return exp.convert(expression.name.startswith(expression.expression.name))

        return expression

    def _is_datetrunc_predicate(self, left: exp.Expr, right: exp.Expr) -> bool:
        return isinstance(left, self.DATETRUNCS) and _is_date_literal(right)

    @annotate_types_on_change
    @catch(ModuleNotFoundError, UnsupportedUnit)
    def simplify_datetrunc(self, expression: exp.Expr) -> exp.Expr:
        """Simplify expressions like `DATE_TRUNC('year', x) >= CAST('2021-01-01' AS DATE)`"""
        comparison = expression.__class__

        if isinstance(expression, self.DATETRUNCS):
            this = expression.this
            trunc_type = extract_type(this)
            date = extract_date(this)
            if date and expression.unit:
                return date_literal(
                    datetime_floor(date, _trunc_unit(expression.unit, self.dialect), self.dialect),
                    trunc_type,
                )
        elif comparison not in self.DATETRUNC_COMPARISONS:
            return expression

        if isinstance(expression, exp.Binary):
            l, r = expression.left, expression.right

            if not self._is_datetrunc_predicate(l, r) or not isinstance(
                l, (exp.DateTrunc, exp.TimestampTrunc)
            ):
                return expression

            trunc_arg = l.this
            unit = _trunc_unit(l.args["unit"], self.dialect)
            date = extract_date(r)

            if not date:
                return expression

            simplified = self.DATETRUNC_BINARY_COMPARISONS[comparison](
                trunc_arg, date, unit, self.dialect, extract_type(r)
            )
            if simplified is None:
                return expression

            return _parenthesize_nested_connector(simplified, expression.parent)

        if isinstance(expression, exp.In):
            l = expression.this
            rs = expression.expressions

            if (
                rs
                and all(self._is_datetrunc_predicate(l, r) for r in rs)
                and isinstance(l, (exp.DateTrunc, exp.TimestampTrunc))
            ):
                unit = _trunc_unit(l.args["unit"], self.dialect)

                ranges = []
                for r in rs:
                    date = extract_date(r)
                    if not date:
                        return expression
                    drange = _datetrunc_range(date, unit, self.dialect)
                    if drange:
                        ranges.append(drange)

                if not ranges:
                    return expression

                ranges = merge_ranges(ranges)
                target_type = extract_type(*rs)

                simplified = exp.or_(
                    *[_datetrunc_eq_expression(l, drange, target_type) for drange in ranges],
                    copy=False,
                )
                return _parenthesize_nested_connector(simplified, expression.parent)

        return expression

    @annotate_types_on_change
    def sort_comparison(self, expression: exp.Expr) -> exp.Expr:
        if expression.__class__ in self.COMPLEMENT_COMPARISONS:
            l, r = expression.this, expression.expression
            l_column = isinstance(l, exp.Column)
            r_column = isinstance(r, exp.Column)
            l_const = _is_constant(l)
            r_const = _is_constant(r)

            if (
                (l_column and not r_column)
                or (r_const and not l_const)
                or isinstance(r, exp.SubqueryPredicate)
            ):
                return expression
            if (r_column and not l_column) or (l_const and not r_const) or (gen(l) > gen(r)):
                return self.INVERSE_COMPARISONS.get(expression.__class__, expression.__class__)(
                    this=r, expression=l
                )
        return expression

    def _flat_simplify(
        self,
        expression: exp.Expr,
        simplifier: t.Callable[[exp.Expr, exp.Expr, exp.Expr], exp.Expr | None],
        root: bool = True,
    ) -> exp.Expr:
        if root or not expression.same_parent:
            operands = []
            queue = deque(expression.flatten(unnest=False))
            size = len(queue)

            # The pairwise scan below is O(n^2). For connectors, a pair only combines if one side
            # is a constant or both are comparisons (see _simplify_connectors / _simplify_comparison);
            # if no operand is combinable the scan is a guaranteed no-op, so return early. This
            # avoids the quadratic blowup on large connectors of inert operands (e.g. a 1000-way OR
            # of ANDs). Non-connector callers (simplify_equality) are unaffected by the type guard.
            if isinstance(expression, exp.Connector) and not any(
                isinstance(op, self.CONNECTOR_COMBINABLE) for op in queue
            ):
                return expression

            while queue:
                a = queue.popleft()

                for b in queue:
                    result = simplifier(expression, a, b)

                    if result and result is not expression:
                        queue.remove(b)
                        queue.appendleft(result)
                        break
                else:
                    operands.append(a)

            if len(operands) < size:
                return functools.reduce(
                    lambda a, b: expression.__class__(this=a, expression=b), operands
                )
        return expression


def gen(expression: exp.Expr, comments: bool = False) -> str:
    """Simple pseudo sql generator for quickly generating sortable and uniq strings.

    Sorting and deduping sql is a necessary step for optimization. Calling the actual
    generator is expensive so we have a bare minimum sql generator here.

    Args:
        expression: the expression to convert into a SQL string.
        comments: whether to include the expression's comments.
    """
    return Gen().gen(expression, comments=comments)


class Gen:
    def __init__(self):
        self.stack = []
        self.sqls = []

    def gen(self, expression: exp.Expr, comments: bool = False) -> str:
        self.stack = [expression]
        self.sqls.clear()
        dispatch = GEN_DISPATCH

        while self.stack:
            node = self.stack.pop()

            if isinstance(node, exp.Expr):
                if comments and node.comments:
                    self.stack.append(f" /*{','.join(node.comments)}*/")

                handler = dispatch.get(node.key)

                if handler is not None:
                    handler(self, node)
                elif isinstance(node, exp.Func):
                    self._function(node)
                else:
                    key = node.key.upper()
                    self.stack.append(f"{key} " if self._args(node) else key)
            elif type(node) is list:
                for n in reversed(node):
                    if n is not None:
                        self.stack.extend((n, ","))
                if node:
                    self.stack.pop()
            else:
                if node is not None:
                    self.sqls.append(str(node))

        return "".join(self.sqls)

    def add_sql(self, e: exp.Add) -> None:
        self._binary(e, " + ")

    def alias_sql(self, e: exp.Alias) -> None:
        self.stack.extend(
            (
                e.args.get("alias"),
                " AS ",
                e.args.get("this"),
            )
        )

    def and_sql(self, e: exp.And) -> None:
        self._binary(e, " AND ")

    def anonymous_sql(self, e: exp.Anonymous) -> None:
        this = e.this
        if isinstance(this, str):
            name = this.upper()
        elif isinstance(this, exp.Identifier):
            name = this.this
            if this.quoted:
                escaped = name.replace('"', '""')
                name = f'"{escaped}"'
            else:
                name = name.upper()
        else:
            raise ValueError(
                f"Anonymous.this expects a str or an Identifier, got '{this.__class__.__name__}'."
            )

        self.stack.extend(
            (
                ")",
                e.expressions,
                "(",
                name,
            )
        )

    def between_sql(self, e: exp.Between) -> None:
        self.stack.extend(
            (
                e.args.get("high"),
                " AND ",
                e.args.get("low"),
                " BETWEEN ",
                e.this,
            )
        )

    def boolean_sql(self, e: exp.Boolean) -> None:
        self.stack.append("TRUE" if e.this else "FALSE")

    def bracket_sql(self, e: exp.Bracket) -> None:
        self.stack.extend(
            (
                "]",
                e.expressions,
                "[",
                e.this,
            )
        )

    def column_sql(self, e: exp.Column) -> None:
        for p in reversed(e.parts):
            self.stack.extend((p, "."))
        self.stack.pop()

    def datatype_sql(self, e: exp.DataType) -> None:
        self._args(e, 1)
        self.stack.append(f"{e.this.name} ")

    def div_sql(self, e: exp.Div) -> None:
        self._binary(e, " / ")

    def dot_sql(self, e: exp.Dot) -> None:
        self._binary(e, ".")

    def eq_sql(self, e: exp.EQ) -> None:
        self._binary(e, " = ")

    def from_sql(self, e: exp.From) -> None:
        self.stack.extend((e.this, "FROM "))

    def gt_sql(self, e: exp.GT) -> None:
        self._binary(e, " > ")

    def gte_sql(self, e: exp.GTE) -> None:
        self._binary(e, " >= ")

    def identifier_sql(self, e: exp.Identifier) -> None:
        if e.quoted:
            escaped = e.this.replace('"', '""')
            self.stack.append(f'"{escaped}"')
        else:
            self.stack.append(e.this)

    def ilike_sql(self, e: exp.ILike) -> None:
        self._binary(e, " NOT ILIKE " if e.args.get("negate") else " ILIKE ")

    def in_sql(self, e: exp.In) -> None:
        self.stack.append(")")
        self._args(e, 1)
        self.stack.extend(
            (
                "(",
                " IN ",
                e.this,
            )
        )

    def intdiv_sql(self, e: exp.IntDiv) -> None:
        self._binary(e, " DIV ")

    def is_sql(self, e: exp.Is) -> None:
        self._binary(e, " IS NOT " if e.args.get("negate") else " IS ")

    def like_sql(self, e: exp.Like) -> None:
        self._binary(e, " NOT Like " if e.args.get("negate") else " Like ")

    def literal_sql(self, e: exp.Literal) -> None:
        if e.is_string:
            escaped = e.this.replace("'", "''")
            self.stack.append(f"'{escaped}'")
        else:
            self.stack.append(e.this)

    def lt_sql(self, e: exp.LT) -> None:
        self._binary(e, " < ")

    def lte_sql(self, e: exp.LTE) -> None:
        self._binary(e, " <= ")

    def mod_sql(self, e: exp.Mod) -> None:
        self._binary(e, " % ")

    def mul_sql(self, e: exp.Mul) -> None:
        self._binary(e, " * ")

    def neg_sql(self, e: exp.Neg) -> None:
        self._unary(e, "-")

    def neq_sql(self, e: exp.NEQ) -> None:
        self._binary(e, " <> ")

    def not_sql(self, e: exp.Not) -> None:
        self._unary(e, "NOT ")

    def null_sql(self, e: exp.Null) -> None:
        self.stack.append("NULL")

    def or_sql(self, e: exp.Or) -> None:
        self._binary(e, " OR ")

    def paren_sql(self, e: exp.Paren) -> None:
        self.stack.extend(
            (
                ")",
                e.this,
                "(",
            )
        )

    def sub_sql(self, e: exp.Sub) -> None:
        self._binary(e, " - ")

    def subquery_sql(self, e: exp.Subquery) -> None:
        self._args(e, 2)
        alias = e.args.get("alias")
        if alias:
            self.stack.append(alias)
        self.stack.extend((")", e.this, "("))

    def table_sql(self, e: exp.Table) -> None:
        self._args(e, 4)
        alias = e.args.get("alias")
        if alias:
            self.stack.append(alias)
        for p in reversed(e.parts):
            self.stack.extend((p, "."))
        self.stack.pop()

    def tablealias_sql(self, e: exp.TableAlias) -> None:
        columns = e.columns

        if columns:
            self.stack.extend((")", columns, "("))

        self.stack.extend((e.this, " AS "))

    def var_sql(self, e: exp.Var) -> None:
        self.stack.append(e.this)

    def _binary(self, e: exp.Binary, op: str) -> None:
        self.stack.extend((e.expression, op, e.this))

    def _unary(self, e: exp.Unary, op: str) -> None:
        self.stack.extend((e.this, op))

    def _function(self, e: exp.Func) -> None:
        self.stack.extend(
            (
                ")",
                list(e.args.values()),
                "(",
                e.sql_name(),
            )
        )

    def _args(self, node: exp.Expr, arg_index: int = 0) -> bool:
        kvs = []
        arg_types = list(node.arg_types)[arg_index:] if arg_index else node.arg_types

        for k in arg_types:
            v = node.args.get(k)

            if v is not None:
                # repr() plain strings so their content can't mimic gen's structural text
                kvs.append([f":{k}", repr(v) if isinstance(v, str) else v])
        if kvs:
            self.stack.append(kvs)
            return True
        return False


def _build_gen_dispatch() -> dict[str, t.Callable[..., None]]:
    # Precompute {expr key -> unbound handler}, mirroring the generator's _build_dispatch, so gen()
    # avoids the per-node f-string + hasattr/getattr lookup (which mypyc can't devirtualize).
    dispatch: dict[str, t.Callable[..., None]] = {}
    for name in dir(Gen):
        if name.endswith("_sql") and not name.startswith("_"):
            dispatch[name[:-4]] = getattr(Gen, name)
    return dispatch


GEN_DISPATCH = _build_gen_dispatch()



from __future__ import annotations

from sqlglot.dialects.postgres import Postgres
from sqlglot.generators.risingwave import RisingWaveGenerator
from sqlglot.parsers.risingwave import RisingWaveParser
from sqlglot.tokens import TokenType


class RisingWave(Postgres):
    NORMALIZE_NOT_NULL = True

    REQUIRES_PARENTHESIZED_STRUCT_ACCESS = True
    SUPPORTS_STRUCT_STAR_EXPANSION = True

    class Tokenizer(Postgres.Tokenizer):
        KEYWORDS = {
            **Postgres.Tokenizer.KEYWORDS,
            "SINK": TokenType.SINK,
            "SOURCE": TokenType.SOURCE,
        }

    Parser = RisingWaveParser

    Generator = RisingWaveGenerator



import ast
import importlib
import pkgutil
import unittest
from pathlib import Path

import sqlglot.typing


def _same_code(f, g):
    if f is g:
        return True
    if not (callable(f) and callable(g) and hasattr(f, "__code__") and hasattr(g, "__code__")):
        return False

    cf, cg = f.__code__, g.__code__
    return (
        cf.co_code == cg.co_code
        and cf.co_names == cg.co_names
        and cf.co_varnames == cg.co_varnames
        and cf.co_consts == cg.co_consts
    )


def _equivalent(child_value, parent_value):
    if set(child_value) != set(parent_value):
        return False

    for key, value in child_value.items():
        parent_entry = parent_value[key]
        if key == "annotator":
            if not _same_code(value, parent_entry):
                return False
        elif value != parent_entry:
            return False

    return True


def _parent_module(source):
    for node in ast.walk(ast.parse(source)):
        if (
            isinstance(node, ast.ImportFrom)
            and node.module
            and node.module.startswith("sqlglot.typing")
            and any(alias.name == "EXPRESSION_METADATA" for alias in node.names)
        ):
            return node.module

    return None


class TestTyping(unittest.TestCase):
    def test_no_redundant_expression_metadata(self):
        """Dialect metadata entries that merely restate the parent module's entry are redundant."""
        package_path = Path(sqlglot.typing.__file__).parent
        redundant = []

        for info in pkgutil.iter_modules([str(package_path)]):
            parent_name = _parent_module((package_path / f"{info.name}.py").read_text("utf-8"))
            if not parent_name:
                continue

            child = importlib.import_module(f"sqlglot.typing.{info.name}").EXPRESSION_METADATA
            parent = importlib.import_module(parent_name).EXPRESSION_METADATA

            for expr_type, child_value in child.items():
                parent_value = parent.get(expr_type)

                # Entries inherited through the `**parent` spread are the same object; only
                # explicitly redefined entries can be redundant
                if (
                    parent_value is not None
                    and child_value is not parent_value
                    and _equivalent(child_value, parent_value)
                ):
                    redundant.append(
                        f"sqlglot.typing.{info.name}: exp.{expr_type.__name__} duplicates "
                        f"the entry inherited from {parent_name}"
                    )

        if redundant:
            self.fail("\n".join(["Redundant entries found:", *redundant]))



from __future__ import annotations

from sqlglot import exp
from sqlglot.typing import EXPRESSION_METADATA

EXPRESSION_METADATA = {
    **EXPRESSION_METADATA,
    **{
        expr_type: {"returns": exp.DType.BIGINT}
        for expr_type in {
            exp.BitLength,
            exp.DateDiff,
            exp.Day,
            exp.DayOfMonth,
            exp.DayOfWeek,
            exp.DayOfWeekIso,
            exp.DayOfYear,
            exp.Extract,
            exp.Hour,
            exp.Length,
            exp.Minute,
            exp.Month,
            exp.Quarter,
            exp.Second,
            exp.Week,
            exp.Year,
        }
    },
    **{
        expr_type: {"returns": exp.DType.INT128}
        for expr_type in {
            exp.CountIf,
            exp.Factorial,
        }
    },
    **{
        expr_type: {"returns": exp.DType.DOUBLE}
        for expr_type in {
            exp.Atan2,
            exp.JarowinklerSimilarity,
            exp.TimeToUnix,
        }
    },
    **{
        expr_type: {"returns": exp.DType.VARCHAR}
        for expr_type in {
            exp.Format,
            exp.Reverse,
            exp.Decode,
        }
    },
    **{
        expr_type: {"returns": exp.DType.VARBINARY}
        for expr_type in {
            exp.Encode,
            exp.Unhex,
        }
    },
    exp.DateBin: {"annotator": lambda self, e: self._annotate_by_args(e, "expression")},
    exp.PercentileDisc: {"annotator": lambda self, e: self._annotate_by_args(e, "this")},
    exp.Localtimestamp: {"returns": exp.DType.TIMESTAMP},
    exp.ToDays: {"returns": exp.DType.INTERVAL},
    exp.TimeFromParts: {"returns": exp.DType.TIME},
}



from __future__ import annotations

import typing as t

from sqlglot import exp, transforms
from sqlglot.dialects.dialect import (
    approx_count_distinct_sql,
    arrow_json_extract_sql,
    rename_func,
    weekstart_unit_to_str,
    inline_array_sql,
    property_sql,
    var_map_sql,
)
from sqlglot.generators.mysql import MySQLGenerator


def _eliminate_between_in_delete(expression: exp.Expr) -> exp.Expr:
    """
    StarRocks doesn't support BETWEEN in DELETE statements, so we convert
    BETWEEN expressions to explicit comparisons.

    https://docs.starrocks.io/docs/sql-reference/sql-statements/table_bucket_part_index/DELETE/#parameters

    Example:
        >>> from sqlglot import parse_one
        >>> expr = parse_one("DELETE FROM t WHERE x BETWEEN 1 AND 10")
        >>> print(_eliminate_between_in_delete(expr).sql(dialect="starrocks"))
        DELETE FROM t WHERE x >= 1 AND x <= 10
    """
    if where := expression.args.get("where"):
        for between in where.find_all(exp.Between):
            between.replace(
                exp.and_(
                    exp.GTE(this=between.this.copy(), expression=between.args["low"]),
                    exp.LTE(this=between.this.copy(), expression=between.args["high"]),
                    copy=False,
                )
            )
    return expression


# https://docs.starrocks.io/docs/sql-reference/sql-functions/spatial-functions/st_distance_sphere/
def st_distance_sphere(self, expression: exp.StDistance) -> str:
    point1 = expression.this
    point2 = expression.expression

    point1_x = self.func("ST_X", point1)
    point1_y = self.func("ST_Y", point1)
    point2_x = self.func("ST_X", point2)
    point2_y = self.func("ST_Y", point2)

    return self.func("ST_Distance_Sphere", point1_x, point1_y, point2_x, point2_y)


class StarRocksGenerator(MySQLGenerator):
    EXCEPT_INTERSECT_SUPPORT_ALL_CLAUSE = False
    JSON_TYPE_REQUIRED_FOR_EXTRACTION = False
    VARCHAR_REQUIRES_SIZE = False
    PARSE_JSON_NAME: str | None = "PARSE_JSON"
    WITH_PROPERTIES_PREFIX = "PROPERTIES"
    UPDATE_STATEMENT_SUPPORTS_FROM = True
    INSERT_OVERWRITE = " OVERWRITE"

    # StarRocks doesn't support "IS TRUE/FALSE" syntax.
    IS_BOOL_ALLOWED = False
    # StarRocks doesn't support renaming a table with a database.
    RENAME_TABLE_WITH_DB = False

    CAST_MAPPING: t.ClassVar[dict[exp.DType, str]] = {}

    TYPE_MAPPING = {
        **MySQLGenerator.TYPE_MAPPING,
        exp.DType.INT128: "LARGEINT",
        exp.DType.TEXT: "STRING",
        exp.DType.TIMESTAMP: "DATETIME",
        exp.DType.TIMESTAMPTZ: "DATETIME",
    }

    SQL_SECURITY_VIEW_LOCATION: t.ClassVar = exp.Properties.Location.POST_SCHEMA

    PROPERTIES_LOCATION = {
        **MySQLGenerator.PROPERTIES_LOCATION,
        exp.PrimaryKey: exp.Properties.Location.POST_SCHEMA,
        exp.UniqueKeyProperty: exp.Properties.Location.POST_SCHEMA,
        exp.RollupProperty: exp.Properties.Location.POST_SCHEMA,
        exp.PartitionedByProperty: exp.Properties.Location.POST_SCHEMA,
    }

    TRANSFORMS = {
        # StarRocks uses the native TRIM(str, chars)/LTRIM/RTRIM function form, not
        # MySQL's TRIM(chars FROM str) syntax, so it falls back to the base generator.
        **{
            k: v for k, v in MySQLGenerator.TRANSFORMS.items() if k not in (exp.DateTrunc, exp.Trim)
        },
        exp.ArgMax: rename_func("MAX_BY"),
        exp.ArgMin: rename_func("MIN_BY"),
        exp.Array: inline_array_sql,
        exp.ArrayAgg: rename_func("ARRAY_AGG"),
        # a <@ b (ArrayContainedBy) is equivalent to ARRAY_CONTAINS_ALL(b, a)
        exp.ArrayContainedBy: lambda self, e: self.func("ARRAY_CONTAINS_ALL", e.expression, e.this),
        exp.ArrayContainsAll: rename_func("ARRAY_CONTAINS_ALL"),
        exp.ArrayFilter: rename_func("ARRAY_FILTER"),
        exp.ArrayToString: rename_func("ARRAY_JOIN"),
        exp.ApproxDistinct: approx_count_distinct_sql,
        exp.CurrentVersion: lambda *_: "CURRENT_VERSION()",
        exp.DateDiff: lambda self, e: self.func(
            "DATE_DIFF", weekstart_unit_to_str(self, e), e.this, e.expression
        ),
        exp.Delete: transforms.preprocess([_eliminate_between_in_delete]),
        exp.Flatten: rename_func("ARRAY_FLATTEN"),
        exp.JSONExtractScalar: arrow_json_extract_sql,
        exp.JSONExtract: arrow_json_extract_sql,
        # Both MAP forms (two-array MAP([keys], [values]) and variadic MAP(k1, v1, ...))
        # generate StarRocks' variadic MAP(k1, v1, k2, v2, ...) constructor
        exp.Map: lambda self, e: var_map_sql(self, e, "MAP"),
        exp.Property: property_sql,
        exp.RegexpLike: rename_func("REGEXP"),
        # Inherited from MySQL, minus operations StarRocks supports natively
        # (QUALIFY, FULL OUTER JOIN, SEMI/ANTI JOIN)
        exp.Select: transforms.preprocess(
            [
                transforms.eliminate_distinct_on,
                transforms.unnest_generate_date_array_using_recursive_cte,
            ]
        ),
        exp.SchemaCommentProperty: lambda self, e: self.naked_property(e),
        exp.SqlSecurityProperty: lambda self, e: f"SECURITY {self.sql(e.this)}",
        exp.StDistance: st_distance_sphere,
        exp.StrToUnix: lambda self, e: self.func("UNIX_TIMESTAMP", e.this, self.format_time(e)),
        exp.TimestampTrunc: lambda self, e: self.func(
            "DATE_TRUNC", weekstart_unit_to_str(self, e), e.this
        ),
        exp.TimeStrToDate: rename_func("TO_DATE"),
        exp.UnixToStr: lambda self, e: self.func("FROM_UNIXTIME", e.this, self.format_time(e)),
        exp.UnixToTime: rename_func("FROM_UNIXTIME"),
        exp.VarMap: lambda self, e: var_map_sql(self, e, "MAP"),
    }

    # https://docs.starrocks.io/docs/sql-reference/sql-statements/keywords/#reserved-keywords
    RESERVED_KEYWORDS = {
        "add",
        "all",
        "alter",
        "analyze",
        "and",
        "array",
        "as",
        "asc",
        "between",
        "bigint",
        "bitmap",
        "both",
        "by",
        "case",
        "char",
        "character",
        "check",
        "collate",
        "column",
        "compaction",
        "convert",
        "create",
        "cross",
        "cube",
        "current_date",
        "current_role",
        "current_time",
        "current_timestamp",
        "current_user",
        "database",
        "databases",
        "decimal",
        "decimalv2",
        "decimal32",
        "decimal64",
        "decimal128",
        "default",
        "deferred",
        "delete",
        "dense_rank",
        "desc",
        "describe",
        "distinct",
        "double",
        "drop",
        "dual",
        "else",
        "except",
        "exists",
        "explain",
        "false",
        "first_value",
        "float",
        "for",
        "force",
        "from",
        "full",
        "function",
        "grant",
        "group",
        "grouping",
        "grouping_id",
        "groups",
        "having",
        "hll",
        "host",
        "if",
        "ignore",
        "immediate",
        "in",
        "index",
        "infile",
        "inner",
        "insert",
        "int",
        "integer",
        "intersect",
        "into",
        "is",
        "join",
        "json",
        "key",
        "keys",
        "kill",
        "lag",
        "largeint",
        "last_value",
        "lateral",
        "lead",
        "left",
        "like",
        "limit",
        "load",
        "localtime",
        "localtimestamp",
        "maxvalue",
        "minus",
        "mod",
        "not",
        "ntile",
        "null",
        "on",
        "or",
        "order",
        "outer",
        "outfile",
        "over",
        "partition",
        "percentile",
        "primary",
        "procedure",
        "qualify",
        "range",
        "rank",
        "read",
        "regexp",
        "release",
        "rename",
        "replace",
        "revoke",
        "right",
        "rlike",
        "row",
        "row_number",
        "rows",
        "schema",
        "schemas",
        "select",
        "set",
        "set_var",
        "show",
        "smallint",
        "system",
        "table",
        "terminated",
        "text",
        "then",
        "tinyint",
        "to",
        "true",
        "union",
        "unique",
        "unsigned",
        "update",
        "use",
        "using",
        "values",
        "varchar",
        "when",
        "where",
        "with",
    }

    def create_sql(self, expression: exp.Create) -> str:
        # Starrocks' primary key is defined outside of the schema, so we need to move it there
        schema = expression.this
        if isinstance(schema, exp.Schema):
            primary_key = schema.find(exp.PrimaryKey)

            if primary_key:
                props = expression.args.get("properties")

                if not props:
                    props = exp.Properties(expressions=[])
                    expression.set("properties", props)

                # Verify if the first one is an engine property. Is true then insert it after the engine,
                # otherwise insert it at the beginning
                engine = props.find(exp.EngineProperty)
                engine_index = (engine.index or 0) if engine else -1
                props.set("expressions", primary_key.pop(), engine_index + 1, overwrite=False)

        return super().create_sql(expression)

    def partitionedbyproperty_sql(self, expression: exp.PartitionedByProperty) -> str:
        this = expression.this
        if isinstance(this, exp.Schema):
            # For MVs, StarRocks needs outer parentheses.
            create = expression.find_ancestor(exp.Create)

            sql = self.expressions(this, flat=True)
            if (create and create.kind == "VIEW") or all(
                isinstance(col, (exp.Column, exp.Identifier)) for col in this.expressions
            ):
                sql = f"({sql})"

            return f"PARTITION BY {sql}"

        return f"PARTITION BY {self.sql(this)}"

    def clusterproperty_sql(self, expression: exp.ClusterProperty) -> str:
        """Generate StarRocks ORDER BY clause for clustering."""
        if expression.this:
            self.unsupported(f"Unsupported CLUSTER BY {self.sql(expression, 'this')}")
            return ""
        expressions = self.expressions(expression, flat=True)
        return f"ORDER BY ({expressions})"

    def refreshtriggerproperty_sql(self, expression: exp.RefreshTriggerProperty) -> str:
        """Generate StarRocks REFRESH clause for materialized views.
        There is a little difference of the syntax between StarRocks and Doris.
        """
        method = self.sql(expression, "method")
        method = f" {method}" if method else ""
        kind = self.sql(expression, "kind")
        kind = f" {kind}" if kind else ""
        starts = self.sql(expression, "starts")
        starts = f" START ({starts})" if starts else ""
        every = self.sql(expression, "every")
        unit = self.sql(expression, "unit")
        every = f" EVERY (INTERVAL {every} {unit})" if every and unit else ""

        return f"REFRESH{method}{kind}{starts}{every}"



from __future__ import annotations

from sqlglot import exp
from sqlglot.parsers.tsql import TSQLParser


class FabricParser(TSQLParser):
    def _parse_create(self) -> exp.Create | exp.Command:
        create = super()._parse_create()

        if isinstance(create, exp.Create):
            # Transform VARCHAR/CHAR without precision to VARCHAR(1)/CHAR(1)
            if create.kind == "TABLE" and isinstance(create.this, exp.Schema):
                for column in create.this.expressions:
                    if isinstance(column, exp.ColumnDef):
                        column_type = column.kind
                        if (
                            isinstance(column_type, exp.DataType)
                            and column_type.this in (exp.DType.VARCHAR, exp.DType.CHAR)
                            and not column_type.expressions
                        ):
                            # Add default precision of 1 to VARCHAR/CHAR without precision
                            # When n isn't specified in a data definition or variable declaration statement, the default length is 1.
                            # https://learn.microsoft.com/en-us/sql/t-sql/data-types/char-and-varchar-transact-sql?view=sql-server-ver17#remarks
                            column_type.set("expressions", [exp.Literal.number("1")])

        return create



"""sqlglot expressions constraints."""

from __future__ import annotations

from sqlglot.expressions.core import Expression, ColumnConstraintKind


class IndexConstraintOption(Expression):
    arg_types = {
        "key_block_size": False,
        "using": False,
        "parser": False,
        "comment": False,
        "visible": False,
        "engine_attr": False,
        "secondary_engine_attr": False,
    }


class Reference(Expression):
    arg_types = {"this": True, "expressions": False, "options": False}


class ColumnConstraint(Expression):
    arg_types = {"this": False, "kind": True}

    @property
    def kind(self) -> ColumnConstraintKind | Reference:
        return self.args["kind"]


class AutoIncrementColumnConstraint(Expression, ColumnConstraintKind):
    pass


class InvisibleColumnConstraint(Expression, ColumnConstraintKind):
    arg_types = {}


class ZeroFillColumnConstraint(ColumnConstraint):
    arg_types = {}


class PeriodForSystemTimeConstraint(Expression, ColumnConstraintKind):
    arg_types = {"this": True, "expression": True}


class CaseSpecificColumnConstraint(Expression, ColumnConstraintKind):
    arg_types = {"not_": True}


class CharacterSetColumnConstraint(Expression, ColumnConstraintKind):
    pass


class CheckColumnConstraint(Expression, ColumnConstraintKind):
    arg_types = {"this": True, "enforced": False}


class AssumeColumnConstraint(Expression, ColumnConstraintKind):
    pass


class ClusteredColumnConstraint(Expression, ColumnConstraintKind):
    pass


class CollateColumnConstraint(Expression, ColumnConstraintKind):
    pass


class CommentColumnConstraint(Expression, ColumnConstraintKind):
    pass


class CompressColumnConstraint(Expression, ColumnConstraintKind):
    arg_types = {"this": False}


class DateFormatColumnConstraint(Expression, ColumnConstraintKind):
    pass


class DefaultColumnConstraint(Expression, ColumnConstraintKind):
    pass


class EncodeColumnConstraint(Expression, ColumnConstraintKind):
    pass


class ExcludeColumnConstraint(Expression, ColumnConstraintKind):
    pass


class EphemeralColumnConstraint(Expression, ColumnConstraintKind):
    arg_types = {"this": False}


class WithOperator(Expression):
    arg_types = {"this": True, "op": True}


class GeneratedAsIdentityColumnConstraint(Expression, ColumnConstraintKind):
    # this: True -> ALWAYS, this: False -> BY DEFAULT
    arg_types = {
        "this": False,
        "expression": False,
        "on_null": False,
        "start": False,
        "increment": False,
        "minvalue": False,
        "maxvalue": False,
        "cycle": False,
        "order": False,
    }


class GeneratedAsRowColumnConstraint(Expression, ColumnConstraintKind):
    arg_types = {"start": False, "hidden": False}


class IndexColumnConstraint(Expression, ColumnConstraintKind):
    arg_types = {
        "this": False,
        "expressions": False,
        "kind": False,
        "index_type": False,
        "options": False,
        "expression": False,  # Clickhouse
        "granularity": False,
    }


class InlineLengthColumnConstraint(Expression, ColumnConstraintKind):
    pass


class NonClusteredColumnConstraint(Expression, ColumnConstraintKind):
    pass


class NotForReplicationColumnConstraint(Expression, ColumnConstraintKind):
    arg_types = {}


class MaskingPolicyColumnConstraint(Expression, ColumnConstraintKind):
    arg_types = {"this": True, "expressions": False}


class NotNullColumnConstraint(Expression, ColumnConstraintKind):
    arg_types = {"allow_null": False}


class OnUpdateColumnConstraint(Expression, ColumnConstraintKind):
    pass


class PrimaryKeyColumnConstraint(Expression, ColumnConstraintKind):
    arg_types = {"desc": False, "options": False}


class TitleColumnConstraint(Expression, ColumnConstraintKind):
    pass


class UniqueColumnConstraint(Expression, ColumnConstraintKind):
    arg_types = {
        "this": False,
        "index_type": False,
        "on_conflict": False,
        "nulls": False,
        "options": False,
    }


class UppercaseColumnConstraint(Expression, ColumnConstraintKind):
    arg_types = {}


class WatermarkColumnConstraint(Expression):
    arg_types = {"this": True, "expression": True}


class PathColumnConstraint(Expression, ColumnConstraintKind):
    pass


class ProjectionPolicyColumnConstraint(Expression, ColumnConstraintKind):
    pass


class ComputedColumnConstraint(Expression, ColumnConstraintKind):
    arg_types = {"this": True, "persisted": False, "not_null": False, "data_type": False}


class InOutColumnConstraint(Expression, ColumnConstraintKind):
    arg_types = {"input_": False, "output": False, "variadic": False}


class Constraint(Expression):
    arg_types = {"this": True, "expressions": True}


class ForeignKey(Expression):
    arg_types = {
        "expressions": False,
        "reference": False,
        "delete": False,
        "update": False,
        "options": False,
    }


class ColumnPrefix(Expression):
    arg_types = {"this": True, "expression": True}


class PrimaryKey(Expression):
    arg_types = {"this": False, "expressions": True, "options": False, "include": False}


# https://docs.databricks.com/aws/en/sql/language-manual/sql-ref-syntax-ddl-create-table-constraint
class TimeseriesKey(Expression):
    pass


class IndexParameters(Expression):
    arg_types = {
        "using": False,
        "include": False,
        "columns": False,
        "with_storage": False,
        "partition_by": False,
        "tablespace": False,
        "where": False,
        "on": False,
    }


class AddConstraint(Expression):
    arg_types = {"expressions": True}



"""sqlglot expressions - JSON functions."""

from __future__ import annotations

from sqlglot.expressions.core import Expression, Func, AggFunc, Binary, Predicate


class CheckJson(Expression, Func):
    arg_types = {"this": True}


class JSONArray(Expression, Func):
    arg_types = {
        "expressions": False,
        "null_handling": False,
        "return_type": False,
        "strict": False,
    }
    is_var_len_args = True


class JSONArrayAgg(Expression, AggFunc):
    arg_types = {
        "this": True,
        "order": False,
        "null_handling": False,
        "return_type": False,
        "strict": False,
    }


class JSONArrayAppend(Expression, Func):
    arg_types = {"this": True, "expressions": True}
    is_var_len_args = True
    _sql_names = ["JSON_ARRAY_APPEND"]


class JSONArrayContains(Expression, Binary, Predicate, Func):
    arg_types = {"this": True, "expression": True, "json_type": False}
    _sql_names = ["JSON_ARRAY_CONTAINS"]


class JSONArrayInsert(Expression, Func):
    arg_types = {"this": True, "expressions": True}
    is_var_len_args = True
    _sql_names = ["JSON_ARRAY_INSERT"]


class JSONBContains(Expression, Binary, Predicate, Func):
    _sql_names = ["JSONB_CONTAINS"]


class JSONBContainsAllTopKeys(Expression, Binary, Predicate, Func):
    pass


class JSONBContainsAnyTopKeys(Expression, Binary, Predicate, Func):
    pass


class JSONBDeleteAtPath(Expression, Binary, Func):
    pass


class JSONBPathExists(Expression, Binary, Predicate, Func):
    pass


class JSONBExists(Expression, Func):
    arg_types = {"this": True, "path": True}
    _sql_names = ["JSONB_EXISTS"]


class JSONBExtract(Expression, Binary, Func):
    _sql_names = ["JSONB_EXTRACT"]


class JSONBExtractScalar(Expression, Binary, Func):
    arg_types = {"this": True, "expression": True, "json_type": False}
    _sql_names = ["JSONB_EXTRACT_SCALAR"]


class JSONBObjectAgg(Expression, AggFunc):
    arg_types = {"this": True, "expression": True}


class JSONBool(Expression, Func):
    pass


class JSONExists(Expression, Func):
    arg_types = {
        "this": True,
        "path": True,
        "passing": False,
        "on_condition": False,
        "from_dcolonqmark": False,
    }


class JSONExtract(Expression, Binary, Func):
    arg_types = {
        "this": True,
        "expression": True,
        "only_json_types": False,
        "expressions": False,
        "variant_extract": False,
        "json_query": False,
        "option": False,
        "quote": False,
        "on_condition": False,
        "requires_json": False,
        "emits": False,
    }
    _sql_names = ["JSON_EXTRACT"]
    is_var_len_args = True

    @property
    def output_name(self) -> str:
        return self.expression.output_name if not self.expressions else ""


class JSONExtractArray(Expression, Func):
    arg_types = {"this": True, "expression": False}
    _sql_names = ["JSON_EXTRACT_ARRAY"]


class JSONExtractScalar(Expression, Binary, Func):
    arg_types = {
        "this": True,
        "expression": True,
        "only_json_types": False,
        "expressions": False,
        "json_type": False,
        "scalar_only": False,
        "json_subtype": False,
    }
    _sql_names = ["JSON_EXTRACT_SCALAR"]
    is_var_len_args = True

    @property
    def output_name(self) -> str:
        return self.expression.output_name


class JSONFormat(Expression, Func):
    arg_types = {"this": False, "options": False, "is_json": False, "to_json": False}
    _sql_names = ["JSON_FORMAT"]


class JSONKeys(Expression, Func):
    arg_types = {"this": True, "expression": False, "expressions": False}
    is_var_len_args = True
    _sql_names = ["JSON_KEYS"]


class JSONKeysAtDepth(Expression, Func):
    arg_types = {"this": True, "expression": False, "mode": False}


class JSONObject(Expression, Func):
    arg_types = {
        "expressions": False,
        "null_handling": False,
        "unique_keys": False,
        "return_type": False,
        "encoding": False,
    }


class JSONObjectAgg(Expression, AggFunc):
    arg_types = {
        "expressions": False,
        "null_handling": False,
        "unique_keys": False,
        "return_type": False,
        "encoding": False,
    }


class JSONRemove(Expression, Func):
    arg_types = {"this": True, "expressions": True}
    is_var_len_args = True
    _sql_names = ["JSON_REMOVE"]


class JSONSet(Expression, Func):
    arg_types = {"this": True, "expressions": True}
    is_var_len_args = True
    _sql_names = ["JSON_SET"]


class JSONStripNulls(Expression, Func):
    arg_types = {
        "this": True,
        "expression": False,
        "include_arrays": False,
        "remove_empty": False,
    }
    _sql_names = ["JSON_STRIP_NULLS"]


class StripNullValue(Expression, Func):
    pass


class JSONTable(Expression, Func):
    arg_types = {
        "this": True,
        "schema": True,
        "path": False,
        "error_handling": False,
        "empty_handling": False,
    }


class JSONType(Expression, Func):
    arg_types = {"this": True, "expression": False}
    _sql_names = ["JSON_TYPE"]


class ObjectId(Expression, Func):
    arg_types = {"this": True, "expression": False}


class ObjectInsert(Expression, Func):
    arg_types = {
        "this": True,
        "key": True,
        "value": True,
        "update_flag": False,
    }


class OpenJSON(Expression, Func):
    arg_types = {"this": True, "path": False, "expressions": False}


class ParseJSON(Expression, Func):
    # BigQuery, Snowflake have PARSE_JSON, Presto has JSON_PARSE
    # Snowflake also has TRY_PARSE_JSON, which is represented using `safe`
    _sql_names = ["PARSE_JSON", "JSON_PARSE"]
    arg_types = {"this": True, "expression": False, "safe": False}



from __future__ import annotations

import typing as t

from collections import defaultdict

from sqlglot import expressions as exp
from sqlglot.helper import find_new_name, seq_get
from sqlglot.optimizer.scope import Scope, traverse_scope

if t.TYPE_CHECKING:
    from sqlglot._typing import E

    FromOrJoin = t.Union[exp.From, exp.Join]


def merge_subqueries(expression: E, leave_tables_isolated: bool = False) -> E:
    """
    Rewrite sqlglot AST to merge derived tables into the outer query.

    This also merges CTEs if they are selected from only once.

    Example:
        >>> import sqlglot
        >>> expression = sqlglot.parse_one("SELECT a FROM (SELECT x.a FROM x) CROSS JOIN y")
        >>> merge_subqueries(expression).sql()
        'SELECT x.a FROM x CROSS JOIN y'

    If `leave_tables_isolated` is True, this will not merge inner queries into outer
    queries if it would result in multiple table selects in a single query:
        >>> expression = sqlglot.parse_one("SELECT a FROM (SELECT x.a FROM x) CROSS JOIN y")
        >>> merge_subqueries(expression, leave_tables_isolated=True).sql()
        'SELECT a FROM (SELECT x.a FROM x) CROSS JOIN y'

    Inspired by https://dev.mysql.com/doc/refman/8.0/en/derived-table-optimization.html

    Args:
        expression (sqlglot.Expr): expression to optimize
        leave_tables_isolated (bool):
    Returns:
        sqlglot.Expr: optimized expression
    """
    # Shared across both passes so the scope tree is only built once, as long as merge_ctes
    # doesn't mutate the AST; if it does, the scopes it was given are no longer valid, so the
    # scope tree needs to be rebuilt before merge_derived_tables runs.
    scopes = traverse_scope(expression)
    expression, merged_ctes = merge_ctes(expression, leave_tables_isolated, scopes=scopes)

    if merged_ctes:
        scopes = traverse_scope(expression)

    expression = merge_derived_tables(expression, leave_tables_isolated, scopes=scopes)
    return expression


# If a derived table has these Select args, it can't be merged
UNMERGABLE_ARGS = set(exp.Select.arg_types) - {
    "expressions",
    "from_",
    "joins",
    "where",
    "order",
    "hint",
}


# Projections in the outer query that are instances of these types can be replaced
# without getting wrapped in parentheses, because the precedence won't be altered.
SAFE_TO_REPLACE_UNWRAPPED = (
    exp.Column,
    exp.EQ,
    exp.Func,
    exp.NEQ,
    exp.Paren,
)


def merge_ctes(
    expression: E,
    leave_tables_isolated: bool = False,
    scopes: list[Scope] | None = None,
) -> tuple[E, bool]:
    # All places where we select from CTEs.
    # We key on the CTE scope so we can detect CTES that are selected from multiple times.
    cte_selections = defaultdict(list)
    for outer_scope in traverse_scope(expression) if scopes is None else scopes:
        for table, inner_scope in outer_scope.selected_sources.values():
            if isinstance(inner_scope, Scope) and inner_scope.is_cte:
                cte_selections[id(inner_scope)].append(
                    (
                        outer_scope,
                        inner_scope,
                        table,
                    )
                )

    merged = False
    singular_cte_selections = [v[0] for k, v in cte_selections.items() if len(v) == 1]
    for outer_scope, inner_scope, table in singular_cte_selections:
        from_or_join = table.find_ancestor(exp.From, exp.Join)
        if not isinstance(from_or_join, (exp.From, exp.Join)):
            continue
        if _mergeable(outer_scope, inner_scope, leave_tables_isolated, from_or_join):
            alias = table.alias_or_name
            _rename_inner_sources(outer_scope, inner_scope, alias)
            _merge_from(outer_scope, inner_scope, table, alias)
            _merge_expressions(outer_scope, inner_scope, alias)
            _merge_order(outer_scope, inner_scope)
            _merge_joins(outer_scope, inner_scope, from_or_join)
            _merge_where(outer_scope, inner_scope, from_or_join)
            _merge_hints(outer_scope, inner_scope)
            _pop_cte(inner_scope)
            outer_scope.clear_cache()
            merged = True
    return expression, merged


def merge_derived_tables(
    expression: E,
    leave_tables_isolated: bool = False,
    scopes: list[Scope] | None = None,
) -> E:
    for outer_scope in traverse_scope(expression) if scopes is None else scopes:
        for subquery in outer_scope.derived_tables:
            from_or_join = subquery.find_ancestor(exp.From, exp.Join)
            if not isinstance(from_or_join, (exp.From, exp.Join)):
                continue
            alias = subquery.alias_or_name
            inner_scope = outer_scope.sources[alias]
            if not isinstance(inner_scope, Scope):
                continue
            if _mergeable(outer_scope, inner_scope, leave_tables_isolated, from_or_join):
                _rename_inner_sources(outer_scope, inner_scope, alias)
                _merge_from(outer_scope, inner_scope, subquery, alias)
                _merge_expressions(outer_scope, inner_scope, alias)
                _merge_order(outer_scope, inner_scope)
                _merge_joins(outer_scope, inner_scope, from_or_join)
                _merge_where(outer_scope, inner_scope, from_or_join)
                _merge_hints(outer_scope, inner_scope)
                outer_scope.clear_cache()

    return expression


def _mergeable(
    outer_scope: Scope,
    inner_scope: Scope,
    leave_tables_isolated: bool,
    from_or_join: FromOrJoin,
) -> bool:
    """
    Return True if `inner_select` can be merged into outer query.
    """
    inner_select = inner_scope.expression.unnest()
    outer = outer_scope.expression
    outer_args = outer.args
    inner_name = from_or_join.alias_or_name

    def _window_projection_blocks_merge(window_aliases: set[str]) -> bool:
        """A window function's result depends on the full row set it sees, so merging the
        subquery into the outer query is unsafe when:
          - the outer query filters or joins (WHERE/JOIN), which changes that row set, or
          - a window column is referenced in an operation that isn't pushed down
            (GROUP BY, ORDER BY, HAVING, aggregate).
        """
        if not window_aliases:
            return False

        if outer_args.get("where") or outer_args.get("joins"):
            return True

        return any(
            column.table == inner_name
            and column.name in window_aliases
            and column.find_ancestor(exp.Group, exp.Order, exp.Having, exp.AggFunc)
            for column in outer_scope.columns
        )

    def _literal_group_unmergeable(number_literal_aliases: set[str]) -> bool:
        """
        A numeric-literal projection referenced in GROUP BY can't be inlined, because a bare
        integer literal is positional. A reference that is itself a top-level GROUP BY item
        can merge as the ordinal of the outer projection that selects it; any other reference,
        e.g., ROLLUP / CUBE / GROUPING SETS, tuples, expressions, etc, blocks the merge, since
        ordinals aren't universally supported there, e.g., Presto / Trino only accept columns.
        """
        group = outer_args.get("group")
        if not group:
            return False

        if not number_literal_aliases:
            return False

        grouped = set()
        top_level_ids = {id(e.unnest()) for e in group.expressions}
        for col in group.find_all(exp.Column):
            if col.table != inner_name or col.name not in number_literal_aliases:
                continue
            if id(col) not in top_level_ids:
                return True
            grouped.add(col.name)

        if not grouped:
            return False

        projected = set()
        for s in t.cast(exp.Select, outer).selects:
            unaliased = s.unalias()
            if isinstance(unaliased, exp.Column) and unaliased.table == inner_name:
                projected.add(unaliased.name)

        return not grouped <= projected

    def _outer_select_joins_on_inner_select_join(projections: dict[str, exp.Expr]) -> bool:
        """
        All columns from the inner select in the ON clause must be from the first FROM table.

        That is, this can be merged:
            SELECT * FROM x JOIN (SELECT y.a AS a FROM y JOIN z) AS q ON x.a = q.a
                                         ^^^           ^
        But this can't:
            SELECT * FROM x JOIN (SELECT z.a AS a FROM y JOIN z) AS q ON x.a = q.a
                                         ^^^                  ^
        """
        if not isinstance(from_or_join, exp.Join):
            return False

        on = from_or_join.args.get("on")
        if not on:
            return False
        selections = [c.name for c in on.find_all(exp.Column) if c.table == inner_name]
        inner_from = inner_scope.expression.args.get("from_")
        if not inner_from:
            return False
        inner_from_table = inner_from.alias_or_name
        return any(
            col.table != inner_from_table
            for selection in selections
            for col in projections[selection].find_all(exp.Column)
        )

    def _is_recursive() -> bool:
        # Recursive CTEs look like this:
        #     WITH RECURSIVE cte AS (
        #       SELECT * FROM x  <-- inner scope
        #       UNION ALL
        #       SELECT * FROM cte  <-- outer scope
        #     )
        cte = inner_scope.expression.parent
        node = outer.parent

        while node:
            if node is cte:
                return True
            node = node.parent
        return False

    def _literal_in_order_by(number_literal_aliases: set[str]) -> bool:
        """A numeric-literal projection under a bare ORDER BY key can't merge (would become positional)."""
        order = outer_args.get("order")
        if not order:
            return False
        ordered = {
            key.name
            for o in order.expressions
            if isinstance(key := o.this.unnest(), exp.Column) and key.table == inner_name
        }
        return bool(number_literal_aliases & ordered)

    if (
        not isinstance(outer, exp.Select)
        or outer.is_star
        or not isinstance(inner_select, exp.Select)
        or any(v for k, v in inner_select.args.items() if k in UNMERGABLE_ARGS)
        or inner_select.args.get("from_") is None
        or outer_scope.pivots
        or (leave_tables_isolated and len(outer_scope.selected_sources) > 1)
        or (isinstance(from_or_join, exp.Join) and inner_select.args.get("joins"))
        or (
            isinstance(from_or_join, exp.Join)
            and inner_select.args.get("where")
            and from_or_join.side in ("FULL", "LEFT", "RIGHT")
        )
        or (
            isinstance(from_or_join, exp.From)
            and inner_select.args.get("where")
            and any(j.side in ("FULL", "RIGHT") for j in outer_args.get("joins", []))
        )
        or (inner_select.args.get("order") and outer_scope.is_union)
        or isinstance(seq_get(inner_select.expressions, 0), exp.QueryTransform)
    ):
        return False

    # Single pass over the projections: replaces the separate AggFunc/Select/Explode and
    # Window tree walks, and precomputes what the checks below need per-projection.
    window_aliases: set[str] = set()
    number_literal_aliases: set[str] = set()
    projections: dict[str, exp.Expr] = {}

    for s in inner_select.selects:
        name = s.alias_or_name
        projections[name] = s
        if s.unalias().is_number:
            number_literal_aliases.add(name)
        for node in s.walk():
            if isinstance(node, (exp.AggFunc, exp.Select, exp.Explode)):
                return False
            if isinstance(node, exp.Window):
                window_aliases.add(name)

    return (
        not _outer_select_joins_on_inner_select_join(projections)
        and not _window_projection_blocks_merge(window_aliases)
        and not _literal_group_unmergeable(number_literal_aliases)
        and not _literal_in_order_by(number_literal_aliases)
        and not (inner_scope.is_cte and _is_recursive())
    )


def _rename_inner_sources(outer_scope: Scope, inner_scope: Scope, alias: str) -> None:
    """
    Renames any sources in the inner query that conflict with names in the outer query.
    """
    inner_taken = set(inner_scope.selected_sources)
    outer_taken = set(outer_scope.selected_sources)
    conflicts = outer_taken.intersection(inner_taken)
    conflicts -= {alias}

    taken = outer_taken.union(inner_taken)

    for conflict in conflicts:
        new_name = find_new_name(taken, conflict)

        source, _ = inner_scope.selected_sources[conflict]
        new_alias = exp.to_identifier(new_name)

        if isinstance(source, exp.Table) and source.alias:
            source.set("alias", exp.TableAlias(this=new_alias))
        elif isinstance(source, exp.Table):
            source.replace(exp.alias_(source, new_alias))
        elif isinstance(source.parent, exp.Subquery):
            source.parent.set("alias", exp.TableAlias(this=new_alias))

        for column in inner_scope.source_columns(conflict):
            column.set("table", exp.to_identifier(new_name))

        inner_scope.rename_source(conflict, new_name)


def _merge_from(
    outer_scope: Scope,
    inner_scope: Scope,
    node_to_replace: exp.Selectable,
    alias: str,
) -> None:
    """
    Merge FROM clause of inner query into outer query.
    """
    new_subquery = inner_scope.expression.args["from_"].this
    new_subquery.set("joins", node_to_replace.args.get("joins"))
    node_to_replace.replace(new_subquery)
    for join_hint in outer_scope.join_hints:
        tables = join_hint.find_all(exp.Table)
        for table in tables:
            if table.alias_or_name == node_to_replace.alias_or_name:
                table.set("this", exp.to_identifier(new_subquery.alias_or_name))
    outer_scope.remove_source(alias)
    outer_scope.add_source(
        new_subquery.alias_or_name, inner_scope.sources[new_subquery.alias_or_name]
    )


def _merge_joins(outer_scope: Scope, inner_scope: Scope, from_or_join: FromOrJoin) -> None:
    """
    Merge JOIN clauses of inner query into outer query.
    """

    new_joins = []

    joins = inner_scope.expression.args.get("joins") or []

    for join in joins:
        new_joins.append(join)
        outer_scope.add_source(join.alias_or_name, inner_scope.sources[join.alias_or_name])

    if new_joins:
        outer_joins = outer_scope.expression.args.get("joins", [])

        # Maintain the join order
        if isinstance(from_or_join, exp.From):
            position = 0
        else:
            position = outer_joins.index(from_or_join) + 1
        outer_joins[position:position] = new_joins

        outer_scope.expression.set("joins", outer_joins)


def _merge_expressions(outer_scope: Scope, inner_scope: Scope, alias: str) -> None:
    """
    Merge projections of inner query into outer query.

    Args:
        outer_scope (sqlglot.optimizer.scope.Scope)
        inner_scope (sqlglot.optimizer.scope.Scope)
        alias (str)
    """
    # Collect all columns that reference the alias of the inner query
    outer_columns = defaultdict(list)
    for column in outer_scope.columns:
        if column.table == alias:
            outer_columns[column.name].append(column)

    group = outer_scope.expression.args.get("group")

    # Replace columns with the projection expression in the inner query
    for expression in inner_scope.expression.expressions:
        projection_name = expression.alias_or_name
        if not projection_name:
            continue
        columns_to_replace = outer_columns.get(projection_name, [])
        if not columns_to_replace:
            continue

        expression = expression.unalias()
        must_wrap_expression = not isinstance(expression, SAFE_TO_REPLACE_UNWRAPPED)

        is_number = expression.is_number
        last = len(columns_to_replace) - 1

        group_ordinal = None
        if is_number and isinstance(outer_scope.expression, exp.Select):
            # Find the ordinal of the outer SELECT that references this inner projection.
            for j, s in enumerate(outer_scope.expression.selects):
                unaliased = s.unalias()
                if (
                    isinstance(unaliased, exp.Column)
                    and unaliased.table == alias
                    and unaliased.name == projection_name
                ):
                    group_ordinal = j + 1
                    break

        for i, column in enumerate(columns_to_replace):
            parent = column.parent

            # A numeric-literal projection can't be inlined into a top-level GROUP BY item
            # (positional context), canonicalize to the projection's ordinal to match
            # qualify. _mergeable guarantees the ordinal exists.
            if is_number and group:
                item = next((e for e in group.expressions if e.unnest() is column), None)
                if item is not None:
                    item.replace(exp.Literal.number(group_ordinal))
                    continue

            # Ensures we don't alter the intended operator precedence if there's additional
            # context surrounding the outer expression (i.e. it's not a simple projection).
            if isinstance(parent, (exp.Unary, exp.Binary)) and must_wrap_expression:
                expression = exp.paren(expression, copy=False)

            # make sure we do not accidentally change the name of the column
            if isinstance(parent, exp.Select) and column.name != expression.name:
                expression = exp.alias_(expression, column.name)

            # Skip the expensive deep copy for the last reference since the inner query
            # is about to be removed, so we can move the expression directly
            column.replace(expression.copy() if i < last else expression)


def _merge_where(outer_scope: Scope, inner_scope: Scope, from_or_join: FromOrJoin) -> None:
    """
    Merge WHERE clause of inner query into outer query.

    Args:
        outer_scope (sqlglot.optimizer.scope.Scope)
        inner_scope (sqlglot.optimizer.scope.Scope)
        from_or_join (exp.From|exp.Join)
    """
    where = inner_scope.expression.args.get("where")
    if not where or not where.this:
        return

    expression = outer_scope.expression

    if isinstance(from_or_join, exp.Join):
        # Merge predicates from an outer join to the ON clause
        # if it only has columns that are already joined
        from_ = expression.args.get("from_")
        sources = {from_.alias_or_name} if from_ else set()

        for join in expression.args["joins"]:
            source = join.alias_or_name
            sources.add(source)
            if source == from_or_join.alias_or_name:
                break

        if exp.column_table_names(where.this) <= sources:
            from_or_join.on(where.this, copy=False)
            from_or_join.set("on", from_or_join.args.get("on"))
            return

    t.cast(exp.Select, expression).where(where.this, copy=False)


def _merge_order(outer_scope: Scope, inner_scope: Scope) -> None:
    """
    Merge ORDER clause of inner query into outer query.

    Args:
        outer_scope (sqlglot.optimizer.scope.Scope)
        inner_scope (sqlglot.optimizer.scope.Scope)
    """
    inner_order = inner_scope.expression.args.get("order")
    if not inner_order:
        return

    if (
        any(
            outer_scope.expression.args.get(arg) for arg in ["group", "distinct", "having", "order"]
        )
        or len(outer_scope.selected_sources) != 1
        or any(expression.find(exp.AggFunc) for expression in outer_scope.expression.expressions)
    ):
        return

    outer_scope.expression.set("order", inner_order)


def _merge_hints(outer_scope: Scope, inner_scope: Scope) -> None:
    inner_scope_hint = inner_scope.expression.args.get("hint")
    if not inner_scope_hint:
        return
    outer_scope_hint = outer_scope.expression.args.get("hint")
    if outer_scope_hint:
        for hint_expression in inner_scope_hint.expressions:
            outer_scope_hint.append("expressions", hint_expression)
    else:
        outer_scope.expression.set("hint", inner_scope_hint)


def _pop_cte(inner_scope: Scope) -> None:
    """
    Remove CTE from the AST.

    Args:
        inner_scope (sqlglot.optimizer.scope.Scope)
    """
    cte = inner_scope.expression.parent
    if not cte:
        return
    with_ = cte.parent
    if not with_:
        return
    if len(with_.expressions) == 1:
        with_.pop()
    else:
        cte.pop()



from __future__ import annotations


from sqlglot import tokens
from sqlglot.dialects.dialect import (
    Dialect,
    NormalizationStrategy,
)
from sqlglot.generators.exasol import ExasolGenerator
from sqlglot.parsers.exasol import DATE_UNITS as DATE_UNITS, ExasolParser
from sqlglot.tokens import TokenType


class Exasol(Dialect):
    # https://docs.exasol.com/db/latest/sql_references/basiclanguageelements.htm#SQLidentifier
    NORMALIZATION_STRATEGY = NormalizationStrategy.UPPERCASE
    # https://docs.exasol.com/db/latest/sql_references/data_types/datatypesoverview.htm
    SUPPORTS_USER_DEFINED_TYPES = False
    # https://docs.exasol.com/db/latest/sql/select.htm
    SUPPORTS_COLUMN_JOIN_MARKS = True
    NULL_ORDERING = "nulls_are_last"
    # https://docs.exasol.com/db/latest/sql_references/literals.htm#StringLiterals
    CONCAT_COALESCE = True

    TIME_MAPPING = {
        "yyyy": "%Y",
        "YYYY": "%Y",
        "yy": "%y",
        "YY": "%y",
        "mm": "%m",
        "MM": "%m",
        "MONTH": "%B",
        "MON": "%b",
        "dd": "%d",
        "DD": "%d",
        "DAY": "%A",
        "DY": "%a",
        "H12": "%I",
        "H24": "%H",
        "HH": "%H",
        "ID": "%u",
        "vW": "%V",
        "IW": "%V",
        "vYYY": "%G",
        "IYYY": "%G",
        "MI": "%M",
        "SS": "%S",
        "uW": "%W",
        "UW": "%U",
        "Z": "%z",
    }

    class Tokenizer(tokens.Tokenizer):
        IDENTIFIERS = ['"', ("[", "]")]
        KEYWORDS = {
            **tokens.Tokenizer.KEYWORDS,
            "USER": TokenType.CURRENT_USER,
            # https://docs.exasol.com/db/latest/sql_references/functions/alphabeticallistfunctions/if.htm
            "ENDIF": TokenType.END,
            "LONG VARCHAR": TokenType.TEXT,
            "REGEXP_LIKE": TokenType.RLIKE,
            "SEPARATOR": TokenType.SEPARATOR,
            "SYSTIMESTAMP": TokenType.SYSTIMESTAMP,
            "MINUS": TokenType.EXCEPT,
        }
        KEYWORDS.pop("DIV")

    Parser = ExasolParser

    Generator = ExasolGenerator



from __future__ import annotations

from sqlglot import exp
from sqlglot.generator import Generator
from sqlglot.generators.postgres import PostgresGenerator


class RisingWaveGenerator(PostgresGenerator):
    LOCKING_READS_SUPPORTED = False
    SUPPORTS_BETWEEN_FLAGS = False

    TRANSFORMS = {
        **PostgresGenerator.TRANSFORMS,
        exp.FileFormatProperty: lambda self, e: f"FORMAT {self.sql(e, 'this')}",
    }

    PROPERTIES_LOCATION = {
        **PostgresGenerator.PROPERTIES_LOCATION,
        exp.FileFormatProperty: exp.Properties.Location.POST_EXPRESSION,
    }

    EXPRESSION_PRECEDES_PROPERTIES_CREATABLES = {"SINK"}

    def computedcolumnconstraint_sql(self, expression: exp.ComputedColumnConstraint) -> str:
        return Generator.computedcolumnconstraint_sql(self, expression)

    def datatype_sql(self, expression: exp.DataType) -> str:
        if expression.is_type(exp.DType.MAP) and len(expression.expressions) == 2:
            key_type, value_type = expression.expressions
            return f"MAP({self.sql(key_type)}, {self.sql(value_type)})"

        return super().datatype_sql(expression)



from __future__ import annotations

from copy import deepcopy
from collections import defaultdict

from sqlglot import exp
from sqlglot.dialects.spark import Spark
from sqlglot.generators.databricks import DatabricksGenerator
from sqlglot.parsers.databricks import DatabricksParser
from sqlglot.tokens import TokenType
from sqlglot.optimizer.annotate_types import TypeAnnotator
from sqlglot.typing.databricks import EXPRESSION_METADATA


class Databricks(Spark):
    SAFE_DIVISION = False
    COPY_PARAMS_ARE_CSV = False
    EXPRESSION_METADATA = EXPRESSION_METADATA.copy()

    COERCES_TO = defaultdict(set, deepcopy(TypeAnnotator.COERCES_TO))
    for text_type in exp.DataType.TEXT_TYPES:
        COERCES_TO[text_type] |= {
            *exp.DataType.NUMERIC_TYPES,
            *exp.DataType.TEMPORAL_TYPES,
            exp.DType.BINARY,
            exp.DType.BOOLEAN,
            exp.DType.INTERVAL,
        }

    class JSONPathTokenizer(Spark.JSONPathTokenizer):
        IDENTIFIERS = ["`", '"']

    class Tokenizer(Spark.Tokenizer):
        KEYWORDS = {
            **Spark.Tokenizer.KEYWORDS,
            "STREAM": TokenType.STREAM,
            "VOID": TokenType.VOID,
        }

    Parser = DatabricksParser

    Generator = DatabricksGenerator



from __future__ import annotations

import typing as t

from sqlglot import exp
from sqlglot.dialects.dialect import Dialect, DialectType

if t.TYPE_CHECKING:
    from sqlglot._typing import E


@t.overload
def normalize_identifiers(
    expression: E, dialect: DialectType = None, store_original_column_identifiers: bool = False
) -> E: ...


@t.overload
def normalize_identifiers(
    expression: str, dialect: DialectType = None, store_original_column_identifiers: bool = False
) -> exp.Identifier: ...


def normalize_identifiers(expression, dialect=None, store_original_column_identifiers=False):
    """
    Normalize identifiers by converting them to either lower or upper case,
    ensuring the semantics are preserved in each case (e.g. by respecting
    case-sensitivity).

    This transformation reflects how identifiers would be resolved by the engine corresponding
    to each SQL dialect, and plays a very important role in the standardization of the AST.

    It's possible to make this a no-op by adding a special comment next to the
    identifier of interest:

        SELECT a /* sqlglot.meta case_sensitive */ FROM table

    In this example, the identifier `a` will not be normalized.

    Note:
        Some dialects (e.g. DuckDB) treat all identifiers as case-insensitive even
        when they're quoted, so in these cases all identifiers are normalized.

    Example:
        >>> import sqlglot
        >>> expression = sqlglot.parse_one('SELECT Bar.A AS A FROM "Foo".Bar')
        >>> normalize_identifiers(expression).sql()
        'SELECT bar.a AS a FROM "Foo".bar'
        >>> normalize_identifiers("foo", dialect="snowflake").sql(dialect="snowflake")
        'FOO'

    Args:
        expression: The expression to transform.
        dialect: The dialect to use in order to decide how to normalize identifiers.
        store_original_column_identifiers: Whether to store the original column identifiers in
            the meta data of the expression in case we want to undo the normalization at a later point.

    Returns:
        The transformed expression.
    """
    dialect = Dialect.get_or_raise(dialect)

    if isinstance(expression, str):
        expression = exp.parse_identifier(expression, dialect=dialect)

    for node in expression.walk(prune=lambda n: bool(n.meta_get("case_sensitive"))):
        if not node.meta_get("case_sensitive"):
            if store_original_column_identifiers and isinstance(node, exp.Column):
                # TODO: This does not handle non-column cases, e.g PARSE_JSON(...).key
                parent = node
                while parent and isinstance(parent.parent, exp.Dot):
                    parent = parent.parent

                node.meta["dot_parts"] = [p.name for p in parent.parts]

            if isinstance(node, exp.Identifier):
                dialect.normalize_identifier(node)

    return expression



from __future__ import annotations

import typing as t

from sqlglot import exp, parser
from sqlglot.trie import new_trie
from sqlglot.dialects.dialect import (
    Dialect,
    build_date_delta,
    build_date_delta_with_interval,
    build_formatted_time,
    isnull_to_is_null,
)
from sqlglot.helper import seq_get
from sqlglot.tokens import TokenType


# All specifiers for time parts (as opposed to date parts)
# https://dev.mysql.com/doc/refman/8.0/en/date-and-time-functions.html#function_date-format
TIME_SPECIFIERS = {"f", "H", "h", "I", "i", "k", "l", "p", "r", "S", "s", "T"}


def _has_time_specifier(date_format: str) -> bool:
    i = 0
    length = len(date_format)

    while i < length:
        if date_format[i] == "%":
            i += 1
            if i < length and date_format[i] in TIME_SPECIFIERS:
                return True
        i += 1
    return False


def _str_to_date(args: list) -> exp.StrToDate | exp.StrToTime:
    mysql_date_format = seq_get(args, 1)
    date_format = Dialect["mysql"].format_time(mysql_date_format)
    this = seq_get(args, 0)

    if mysql_date_format and _has_time_specifier(mysql_date_format.name):
        return exp.StrToTime(this=this, format=date_format)

    return exp.StrToDate(this=this, format=date_format)


def _show_parser(*args: t.Any, **kwargs: t.Any) -> t.Callable[[MySQLParser], exp.Show]:
    def _parse(self: MySQLParser) -> exp.Show:
        return self._parse_show_mysql(*args, **kwargs)

    return _parse


class MySQLParser(parser.Parser):
    NO_PAREN_FUNCTIONS = {
        **parser.Parser.NO_PAREN_FUNCTIONS,
        TokenType.LOCALTIME: exp.Localtime,
        TokenType.LOCALTIMESTAMP: exp.Localtimestamp,
    }

    ID_VAR_TOKENS = parser.Parser.ID_VAR_TOKENS - {TokenType.STRAIGHT_JOIN}

    FUNC_TOKENS = {
        *parser.Parser.FUNC_TOKENS,
        TokenType.DATABASE,
        TokenType.MOD,
        TokenType.SCHEMA,
        TokenType.VALUES,
        TokenType.CHARACTER_SET,
    }

    CONJUNCTION = {
        **parser.Parser.CONJUNCTION,
        TokenType.DAMP: exp.And,
        TokenType.XOR: exp.Xor,
    }

    DISJUNCTION = {
        **parser.Parser.DISJUNCTION,
        TokenType.DPIPE: exp.Or,
    }

    TABLE_ALIAS_TOKENS = (
        (parser.Parser.TABLE_ALIAS_TOKENS | {TokenType.ANTI, TokenType.SEMI})
        - parser.Parser.TABLE_INDEX_HINT_TOKENS
        - {TokenType.STRAIGHT_JOIN}
    )

    RANGE_PARSERS = {
        **parser.Parser.RANGE_PARSERS,
        TokenType.MEMBER_OF: lambda self, this: self.expression(
            exp.JSONArrayContains(this=this, expression=self._parse_wrapped(self._parse_expression))
        ),
    }

    FUNCTIONS = {
        **parser.Parser.FUNCTIONS,
        "BIT_AND": exp.BitwiseAndAgg.from_arg_list,
        "BIT_OR": exp.BitwiseOrAgg.from_arg_list,
        "BIT_XOR": exp.BitwiseXorAgg.from_arg_list,
        "BIT_COUNT": exp.BitwiseCount.from_arg_list,
        "CONVERT_TZ": lambda args: exp.ConvertTimezone(
            source_tz=seq_get(args, 1), target_tz=seq_get(args, 2), timestamp=seq_get(args, 0)
        ),
        "CURDATE": exp.CurrentDate.from_arg_list,
        "CURTIME": exp.CurrentTime.from_arg_list,
        "DATE": lambda args: exp.TsOrDsToDate(this=seq_get(args, 0)),
        "DATEDIFF": lambda args: exp.DateDiff(
            this=seq_get(args, 0), expression=seq_get(args, 1), date_part_boundary=True
        ),
        "DATE_ADD": build_date_delta_with_interval(exp.DateAdd),
        "DATE_FORMAT": lambda args: exp.TimeToStr(
            this=exp.TsOrDsToTimestamp(this=seq_get(args, 0)),
            format=Dialect["mysql"].format_time(seq_get(args, 1)),
        ),
        "DATE_SUB": build_date_delta_with_interval(exp.DateSub),
        "DAY": lambda args: exp.Day(this=exp.TsOrDsToDate(this=seq_get(args, 0))),
        "DAYOFMONTH": lambda args: exp.DayOfMonth(this=exp.TsOrDsToDate(this=seq_get(args, 0))),
        "DAYOFWEEK": lambda args: exp.DayOfWeek(this=exp.TsOrDsToDate(this=seq_get(args, 0))),
        "DAYOFYEAR": lambda args: exp.DayOfYear(this=exp.TsOrDsToDate(this=seq_get(args, 0))),
        "FORMAT": exp.NumberToStr.from_arg_list,
        "FROM_UNIXTIME": build_formatted_time(exp.UnixToTime),
        "ISNULL": isnull_to_is_null,
        "LENGTH": lambda args: exp.Length(this=seq_get(args, 0), binary=True),
        "MAKETIME": exp.TimeFromParts.from_arg_list,
        "MONTH": lambda args: exp.Month(this=exp.TsOrDsToDate(this=seq_get(args, 0))),
        "MONTHNAME": lambda args: exp.TimeToStr(
            this=exp.TsOrDsToDate(this=seq_get(args, 0)),
            format=exp.Literal.string("%B"),
        ),
        "SCHEMA": exp.CurrentSchema.from_arg_list,
        "DATABASE": exp.CurrentSchema.from_arg_list,
        "STR_TO_DATE": _str_to_date,
        "TIMESTAMPDIFF": build_date_delta(exp.TimestampDiff),
        "TO_DAYS": lambda args: exp.paren(
            exp.DateDiff(
                this=exp.TsOrDsToDate(this=seq_get(args, 0)),
                expression=exp.TsOrDsToDate(this=exp.Literal.string("0000-01-01")),
                unit=exp.var("DAY"),
            )
            + 1
        ),
        "VERSION": exp.CurrentVersion.from_arg_list,
        "WEEK": lambda args: exp.Week(
            this=exp.TsOrDsToDate(this=seq_get(args, 0)), mode=seq_get(args, 1)
        ),
        "WEEKOFYEAR": lambda args: exp.WeekOfYear(this=exp.TsOrDsToDate(this=seq_get(args, 0))),
        "YEAR": lambda args: exp.Year(this=exp.TsOrDsToDate(this=seq_get(args, 0))),
    }

    FUNCTION_PARSERS = {
        **parser.Parser.FUNCTION_PARSERS,
        "GROUP_CONCAT": lambda self: self._parse_group_concat(),
        # https://dev.mysql.com/doc/refman/5.7/en/miscellaneous-functions.html#function_values
        "VALUES": lambda self: self.expression(
            exp.Anonymous(this="VALUES", expressions=[self._parse_id_var()])
        ),
        "JSON_VALUE": lambda self: self._parse_json_value(),
        "SUBSTR": lambda self: self._parse_substring(),
    }

    STATEMENT_PARSERS = {
        **parser.Parser.STATEMENT_PARSERS,
        TokenType.SHOW: lambda self: self._parse_show(),
    }

    SHOW_PARSERS = {
        "BINARY LOGS": _show_parser("BINARY LOGS"),
        "MASTER LOGS": _show_parser("BINARY LOGS"),
        "BINLOG EVENTS": _show_parser("BINLOG EVENTS"),
        "CHARACTER SET": _show_parser("CHARACTER SET"),
        "CHARSET": _show_parser("CHARACTER SET"),
        "COLLATION": _show_parser("COLLATION"),
        "FULL COLUMNS": _show_parser("COLUMNS", target="FROM", full=True),
        "COLUMNS": _show_parser("COLUMNS", target="FROM"),
        "CREATE DATABASE": _show_parser("CREATE DATABASE", target=True),
        "CREATE EVENT": _show_parser("CREATE EVENT", target=True),
        "CREATE FUNCTION": _show_parser("CREATE FUNCTION", target=True),
        "CREATE PROCEDURE": _show_parser("CREATE PROCEDURE", target=True),
        "CREATE TABLE": _show_parser("CREATE TABLE", target=True),
        "CREATE TRIGGER": _show_parser("CREATE TRIGGER", target=True),
        "CREATE VIEW": _show_parser("CREATE VIEW", target=True),
        "DATABASES": _show_parser("DATABASES"),
        "SCHEMAS": _show_parser("DATABASES"),
        "ENGINE": _show_parser("ENGINE", target=True),
        "STORAGE ENGINES": _show_parser("ENGINES"),
        "ENGINES": _show_parser("ENGINES"),
        "ERRORS": _show_parser("ERRORS"),
        "EVENTS": _show_parser("EVENTS"),
        "FUNCTION CODE": _show_parser("FUNCTION CODE", target=True),
        "FUNCTION STATUS": _show_parser("FUNCTION STATUS"),
        "GRANTS": _show_parser("GRANTS", target="FOR"),
        "INDEX": _show_parser("INDEX", target="FROM"),
        "MASTER STATUS": _show_parser("MASTER STATUS"),
        "OPEN TABLES": _show_parser("OPEN TABLES"),
        "PLUGINS": _show_parser("PLUGINS"),
        "PROCEDURE CODE": _show_parser("PROCEDURE CODE", target=True),
        "PROCEDURE STATUS": _show_parser("PROCEDURE STATUS"),
        "PRIVILEGES": _show_parser("PRIVILEGES"),
        "FULL PROCESSLIST": _show_parser("PROCESSLIST", full=True),
        "PROCESSLIST": _show_parser("PROCESSLIST"),
        "PROFILE": _show_parser("PROFILE"),
        "PROFILES": _show_parser("PROFILES"),
        "RELAYLOG EVENTS": _show_parser("RELAYLOG EVENTS"),
        "REPLICAS": _show_parser("REPLICAS"),
        "SLAVE HOSTS": _show_parser("REPLICAS"),
        "REPLICA STATUS": _show_parser("REPLICA STATUS"),
        "SLAVE STATUS": _show_parser("REPLICA STATUS"),
        "GLOBAL STATUS": _show_parser("STATUS", global_=True),
        "SESSION STATUS": _show_parser("STATUS"),
        "STATUS": _show_parser("STATUS"),
        "TABLE STATUS": _show_parser("TABLE STATUS"),
        "FULL TABLES": _show_parser("TABLES", full=True),
        "TABLES": _show_parser("TABLES"),
        "TRIGGERS": _show_parser("TRIGGERS"),
        "GLOBAL VARIABLES": _show_parser("VARIABLES", global_=True),
        "SESSION VARIABLES": _show_parser("VARIABLES"),
        "VARIABLES": _show_parser("VARIABLES"),
        "WARNINGS": _show_parser("WARNINGS"),
    }

    PROPERTY_PARSERS = {
        **parser.Parser.PROPERTY_PARSERS,
        "LOCK": lambda self: self._parse_property_assignment(exp.LockProperty),
        "PARTITION BY": lambda self: self._parse_partition_property(),
    }

    SET_PARSERS = {
        **parser.Parser.SET_PARSERS,
        "PERSIST": lambda self: self._parse_set_item_assignment("PERSIST"),
        "PERSIST_ONLY": lambda self: self._parse_set_item_assignment("PERSIST_ONLY"),
        "CHARACTER SET": lambda self: self._parse_set_item_charset("CHARACTER SET"),
        "CHARSET": lambda self: self._parse_set_item_charset("CHARACTER SET"),
        "NAMES": lambda self: self._parse_set_item_names(),
    }

    SHOW_TRIE = new_trie(key.split(" ") for key in SHOW_PARSERS)
    SET_TRIE = new_trie(key.split(" ") for key in SET_PARSERS)

    CONSTRAINT_PARSERS = {
        **parser.Parser.CONSTRAINT_PARSERS,
        "FULLTEXT": lambda self: self._parse_index_constraint(kind="FULLTEXT"),
        "INDEX": lambda self: self._parse_index_constraint(),
        "KEY": lambda self: self._parse_index_constraint(),
        "SPATIAL": lambda self: self._parse_index_constraint(kind="SPATIAL"),
        "ZEROFILL": lambda self: self.expression(exp.ZeroFillColumnConstraint()),
        "INVISIBLE": lambda self: self.expression(exp.InvisibleColumnConstraint()),
    }

    ALTER_PARSERS = {
        **parser.Parser.ALTER_PARSERS,
        "CHANGE": lambda self: self._parse_alter_table_modify(rename=True),
        "MODIFY": lambda self: self._parse_alter_table_modify(),
        "AUTO_INCREMENT": lambda self: self._parse_property_assignment(exp.AutoIncrementProperty),
    }

    ALTER_ALTER_PARSERS = {
        **parser.Parser.ALTER_ALTER_PARSERS,
        "INDEX": lambda self: self._parse_alter_table_alter_index(),
    }

    SCHEMA_UNNAMED_CONSTRAINTS = {
        *parser.Parser.SCHEMA_UNNAMED_CONSTRAINTS,
        "FULLTEXT",
        "INDEX",
        "KEY",
        "SPATIAL",
    }

    PROFILE_TYPES: t.ClassVar[parser.OPTIONS_TYPE] = {
        **dict.fromkeys(("ALL", "CPU", "IPC", "MEMORY", "SOURCE", "SWAPS"), tuple()),
        "BLOCK": ("IO",),
        "CONTEXT": ("SWITCHES",),
        "PAGE": ("FAULTS",),
    }

    TYPE_TOKENS = {
        *parser.Parser.TYPE_TOKENS,
        TokenType.SET,
    }

    ENUM_TYPE_TOKENS = {
        *parser.Parser.ENUM_TYPE_TOKENS,
        TokenType.SET,
    }

    # SELECT [ ALL | DISTINCT | DISTINCTROW ] [ <OPERATION_MODIFIERS> ]
    OPERATION_MODIFIERS = {
        "HIGH_PRIORITY",
        "STRAIGHT_JOIN",
        "SQL_SMALL_RESULT",
        "SQL_BIG_RESULT",
        "SQL_BUFFER_RESULT",
        "SQL_NO_CACHE",
        "SQL_CALC_FOUND_ROWS",
    }

    LOG_DEFAULTS_TO_LN = True
    STRING_ALIASES = True
    VALUES_FOLLOWED_BY_PAREN = False
    SUPPORTS_PARTITION_SELECTION = True

    def _parse_range(self, this: exp.Expr | None = None) -> exp.Expr | None:
        this = this or self._parse_bitwise()

        if self._match_text_seq("SOUNDS", "LIKE"):
            this = self.expression(
                exp.EQ(
                    this=self.expression(exp.Soundex(this=this)),
                    expression=self.expression(exp.Soundex(this=self._parse_bitwise())),
                )
            )
            # MySQL evaluates = and IS left to right because they have the same precedence.
            if self._match(TokenType.IS, advance=False):
                this = self.expression(exp.Paren(this=this))

        return super()._parse_range(this)

    def _parse_alter_table_rename(self):
        if self._match_texts(("INDEX", "KEY")):
            old = self._parse_field(any_token=True)
            self._match_text_seq("TO")
            new = self._parse_field(any_token=True)
            return self.expression(exp.RenameIndex(this=old, to=new))
        return super()._parse_alter_table_rename()

    def _parse_alter_drop_action(self) -> exp.Expr | None:
        if self._match_pair(TokenType.DROP, TokenType.PRIMARY_KEY):
            return self.expression(exp.DropPrimaryKey())
        return super()._parse_alter_drop_action()

    def _parse_alter_table_modify(self, rename: bool = False) -> exp.Expr | None:
        # MODIFY [COLUMN]            col_name      column_definition [FIRST | AFTER col_name]
        # CHANGE [COLUMN] old_col_name new_col_name column_definition [FIRST | AFTER col_name]
        self._match(TokenType.COLUMN)

        column = self._parse_field(any_token=True)
        if column is None:
            return None

        rename_from = None
        if rename:
            rename_from = column
            column = self._parse_field(any_token=True)
            if column is None:
                return None

        column_def = self._parse_column_def(column)
        if not isinstance(column_def, exp.ColumnDef):
            return None

        return self.expression(exp.ModifyColumn(this=column_def, rename_from=rename_from))

    def _parse_generated_as_identity(
        self,
    ) -> (
        exp.GeneratedAsIdentityColumnConstraint
        | exp.ComputedColumnConstraint
        | exp.GeneratedAsRowColumnConstraint
    ):
        this = super()._parse_generated_as_identity()

        if self._match_texts(("STORED", "VIRTUAL")):
            persisted = self._prev.text.upper() == "STORED"

            if isinstance(this, exp.ComputedColumnConstraint):
                this.set("persisted", persisted)
            elif isinstance(this, exp.GeneratedAsIdentityColumnConstraint):
                this = self.expression(
                    exp.ComputedColumnConstraint(this=this.expression, persisted=persisted)
                )

        return this

    def _parse_primary_key_part(self) -> exp.Expr | None:
        this = self._parse_id_var()
        if not self._match(TokenType.L_PAREN):
            return this

        expression = self._parse_number()
        self._match_r_paren()
        return self.expression(exp.ColumnPrefix(this=this, expression=expression))

    def _parse_index_constraint(self, kind: str | None = None) -> exp.IndexColumnConstraint:
        if kind:
            self._match_texts(("INDEX", "KEY"))

        this = self._parse_id_var(any_token=False)
        index_type = self._match(TokenType.USING) and self._advance_any() and self._prev.text
        expressions = self._parse_wrapped_csv(self._parse_ordered)

        options = []
        while True:
            if self._match_text_seq("KEY_BLOCK_SIZE"):
                self._match(TokenType.EQ)
                opt = exp.IndexConstraintOption(key_block_size=self._parse_number())
            elif self._match(TokenType.USING):
                opt = exp.IndexConstraintOption(using=self._advance_any() and self._prev.text)
            elif self._match_text_seq("WITH", "PARSER"):
                opt = exp.IndexConstraintOption(parser=self._parse_var(any_token=True))
            elif self._match(TokenType.COMMENT):
                opt = exp.IndexConstraintOption(comment=self._parse_string())
            elif self._match_text_seq("VISIBLE"):
                opt = exp.IndexConstraintOption(visible=True)
            elif self._match_text_seq("INVISIBLE"):
                opt = exp.IndexConstraintOption(visible=False)
            elif self._match_text_seq("ENGINE_ATTRIBUTE"):
                self._match(TokenType.EQ)
                opt = exp.IndexConstraintOption(engine_attr=self._parse_string())
            elif self._match_text_seq("SECONDARY_ENGINE_ATTRIBUTE"):
                self._match(TokenType.EQ)
                opt = exp.IndexConstraintOption(secondary_engine_attr=self._parse_string())
            else:
                opt = None

            if not opt:
                break

            options.append(opt)

        return self.expression(
            exp.IndexColumnConstraint(
                this=this,
                expressions=expressions,
                kind=kind,
                index_type=index_type,
                options=options,
            )
        )

    def _parse_show_mysql(
        self,
        this: str,
        target: bool | str = False,
        full: bool | None = None,
        global_: bool | None = None,
    ) -> exp.Show:
        json = self._match_text_seq("JSON")

        if target:
            if isinstance(target, str):
                self._match_text_seq(*target.split(" "))
            target_id = self._parse_id_var()
        else:
            target_id = None

        index = self._index
        if self._match_text_seq("IN"):
            log = self._parse_string()
            if log is None:
                self._retreat(index)
        else:
            log = None

        if this in ("BINLOG EVENTS", "RELAYLOG EVENTS"):
            position = self._parse_number() if self._match_text_seq("FROM") else None
            db = None
        else:
            position = None
            db = None

            if self._match(TokenType.FROM) or self._match_text_seq("IN"):
                db = self._parse_id_var()
            elif self._match(TokenType.DOT):
                db = target_id
                target_id = self._parse_id_var()

        channel = self._parse_id_var() if self._match_text_seq("FOR", "CHANNEL") else None

        like = self._parse_string() if self._match_text_seq("LIKE") else None
        where = self._parse_where()

        if this == "PROFILE":
            types = self._parse_csv(lambda: self._parse_var_from_options(self.PROFILE_TYPES))
            query = self._parse_number() if self._match_text_seq("FOR", "QUERY") else None
            offset = self._parse_number() if self._match_text_seq("OFFSET") else None
            limit = self._parse_number() if self._match_text_seq("LIMIT") else None
        else:
            types, query = None, None
            offset, limit = self._parse_oldstyle_limit()

        mutex = True if self._match_text_seq("MUTEX") else None
        mutex = False if self._match_text_seq("STATUS") else mutex

        for_table = self._parse_id_var() if self._match_text_seq("FOR", "TABLE") else None
        for_group = self._parse_string() if self._match_text_seq("FOR", "GROUP") else None
        for_user = self._parse_string() if self._match_text_seq("FOR", "USER") else None
        for_role = self._parse_string() if self._match_text_seq("FOR", "ROLE") else None
        into_outfile = self._parse_string() if self._match_text_seq("INTO", "OUTFILE") else None

        return self.expression(
            exp.Show(
                this=this,
                target=target_id,
                full=full,
                log=log,
                position=position,
                db=db,
                channel=channel,
                like=like,
                where=where,
                types=types,
                query=query,
                offset=offset,
                limit=limit,
                mutex=mutex,
                for_table=for_table,
                for_group=for_group,
                for_user=for_user,
                for_role=for_role,
                into_outfile=into_outfile,
                json=json,
                global_=global_,
            )
        )

    def _parse_oldstyle_limit(
        self,
    ) -> tuple[exp.Expr | None, exp.Expr | None]:
        limit = None
        offset = None
        if self._match_text_seq("LIMIT"):
            parts = self._parse_csv(self._parse_number)
            if len(parts) == 1:
                limit = parts[0]
            elif len(parts) == 2:
                limit = parts[1]
                offset = parts[0]

        return offset, limit

    def _parse_set_item_charset(self, kind: str) -> exp.Expr:
        this = self._parse_string() or self._parse_unquoted_field()
        return self.expression(exp.SetItem(this=this, kind=kind))

    def _parse_charset_name(self) -> exp.Expr | None:
        """
        Preserve quoting when a charset name has characters that require it (e.g. spaces, as allowed
        for custom XML-registered charsets). Safe names unwrap to a bare Var so roundtrips remain minimal.
        """
        identifier = self._parse_identifier()
        if identifier:
            return (
                exp.Var(this=name)
                if exp.SAFE_IDENTIFIER_RE.match(name := identifier.name)
                else identifier
            )
        return self._parse_var(tokens={TokenType.BINARY})

    def _parse_set_item_names(self) -> exp.Expr:
        charset = self._parse_string() or self._parse_unquoted_field()
        if self._match_text_seq("COLLATE"):
            collate = self._parse_string() or self._parse_unquoted_field()
        else:
            collate = None

        return self.expression(exp.SetItem(this=charset, collate=collate, kind="NAMES"))

    def _parse_type(
        self, parse_interval: bool = True, fallback_to_identifier: bool = False
    ) -> exp.Expr | None:
        # mysql binary is special and can work anywhere, even in order by operations
        # it operates like a no paren func
        if self._match(TokenType.BINARY, advance=False):
            data_type = self._parse_types(check_func=True, allow_identifiers=False)

            if isinstance(data_type, exp.DataType):
                return self.expression(exp.Cast(this=self._parse_column(), to=data_type))

        return super()._parse_type(
            parse_interval=parse_interval, fallback_to_identifier=fallback_to_identifier
        )

    def _parse_alter_table_alter_index(self) -> exp.AlterIndex:
        index = self._parse_field(any_token=True)

        if self._match_text_seq("VISIBLE"):
            visible = True
        elif self._match_text_seq("INVISIBLE"):
            visible = False
        else:
            visible = None

        return self.expression(exp.AlterIndex(this=index, visible=visible))

    def _parse_partition_property(
        self,
    ) -> exp.Expr | None | list[exp.Expr]:
        partition_cls: type[exp.Expr] | None = None
        value_parser = None

        if self._match_text_seq("RANGE"):
            partition_cls = exp.PartitionByRangeProperty
            value_parser = self._parse_partition_range_value
        elif self._match_text_seq("LIST"):
            partition_cls = exp.PartitionByListProperty
            value_parser = self._parse_partition_list_value

        if not partition_cls or not value_parser:
            return None

        partition_expressions = self._parse_wrapped_csv(self._parse_assignment)

        # For Doris and Starrocks
        if not self._match_text_seq("(", "PARTITION", advance=False):
            return partition_expressions

        create_expressions = self._parse_wrapped_csv(value_parser)

        return self.expression(
            partition_cls(
                partition_expressions=partition_expressions, create_expressions=create_expressions
            )
        )

    def _parse_partition_range_value(self) -> exp.Expr | None:
        self._match_text_seq("PARTITION")
        name = self._parse_id_var()

        if not self._match_text_seq("VALUES", "LESS", "THAN"):
            return name

        values = self._parse_wrapped_csv(self._parse_expression)

        if (
            len(values) == 1
            and isinstance(values[0], exp.Column)
            and values[0].name.upper() == "MAXVALUE"
        ):
            values = [exp.var("MAXVALUE")]

        part_range = self.expression(exp.PartitionRange(this=name, expressions=values))
        return self.expression(exp.Partition(expressions=[part_range]))

    def _parse_partition_list_value(self) -> exp.Partition:
        self._match_text_seq("PARTITION")
        name = self._parse_id_var()
        self._match_text_seq("VALUES", "IN")
        values = self._parse_wrapped_csv(self._parse_expression)
        part_list = self.expression(exp.PartitionList(this=name, expressions=values))
        return self.expression(exp.Partition(expressions=[part_list]))

    def _parse_primary_key(
        self,
        wrapped_optional: bool = False,
        in_props: bool = False,
        named_primary_key: bool = False,
    ) -> exp.PrimaryKeyColumnConstraint | exp.PrimaryKey:
        return super()._parse_primary_key(
            wrapped_optional=wrapped_optional, in_props=in_props, named_primary_key=True
        )



from tests.dialects.test_dialect import Validator


class TestRisingWave(Validator):
    dialect = "risingwave"
    maxDiff = None

    def test_risingwave(self):
        self.validate_all(
            "SELECT a FROM tbl",
            read={
                "": "SELECT a FROM tbl FOR UPDATE",
            },
        )
        self.validate_identity(
            "CREATE SOURCE from_kafka (*, gen_i32_field INT AS int32_field + 2, gen_i64_field INT AS int64_field + 2, WATERMARK FOR time_col AS time_col - INTERVAL '5 SECOND') INCLUDE header foo VARCHAR AS myheader INCLUDE key AS mykey WITH (connector='kafka', topic='my_topic') FORMAT PLAIN ENCODE PROTOBUF (A=1, B=2) KEY ENCODE PROTOBUF (A=3, B=4)"
        )
        self.validate_identity(
            "CREATE SINK my_sink AS SELECT * FROM A WITH (connector='kafka', topic='my_topic') FORMAT PLAIN ENCODE PROTOBUF (A=1, B=2) KEY ENCODE PROTOBUF (A=3, B=4)"
        )
        self.validate_identity(
            "WITH t1 AS MATERIALIZED (SELECT 1), t2 AS NOT MATERIALIZED (SELECT 2) SELECT * FROM t1, t2"
        )

    def test_datatypes(self):
        self.validate_identity("SELECT CAST(NULL AS MAP(VARCHAR, INT)) AS map_column")

        self.validate_identity(
            "SELECT NULL::MAP<VARCHAR, INT> AS map_column",
            "SELECT CAST(NULL AS MAP(VARCHAR, INT)) AS map_column",
        )

        self.validate_identity("CREATE TABLE t (map_col MAP(VARCHAR, INT))")

        self.validate_identity(
            "CREATE TABLE t (map_col MAP<VARCHAR, INT>)",
            "CREATE TABLE t (map_col MAP(VARCHAR, INT))",
        )
